""" This module contains simulation function implementations of the DTLZ test
suite, as described in:

 * Deb, Thiele, Laumanns, and Zitzler. "Scalable test problems for
   evolutionary multiobjective optimization" in Evolutionary
   Multiobjective Optimization, Theoretical Advances and Applications,
   Ch. 6 (pp. 105--145). Springer-Verlag, London, UK, 2005. Abraham,
   Jain, and Goldberg (Eds).

When run with default settings, the outputs of ``dtlz{1-7}_sim`` have
been confirmed against the outputs from the corresponding problems
in ``pymoo`` up to 8 decimal places of precision.

 * Blank and Deb. "pymoo: Multi-Objective Optimization in Python."
   IEEE Access 8 (pp. 89497--89509). 2020.


One drawback of the original DTLZ problems is that their global minima
(Pareto points) always corresponded to design points that satisfy

x_i = 0.5, for i = number of objectives, ..., number of design points

or

x_i = 0, for i = number of objectives, ..., number of design points.

This was appropriate for testing evolutionary algorithms, but for many
deterministic algorithms, these solutions may represent either the
best- or worst-case scenarios.

To make these problems applicable for deterministic algorithms, the
solution sets must be configurable offset by a user-specified amount,
as proposed in:

 * Chang. Mathematical Software for Multiobjective Optimization Problems.
   Ph.D. dissertation, Virginia Tech, Dept. of Computer Science, 2020.

For the problems DTLZ8 and DTLZ9, only objective outputs are given
by the simulation function herein. To fully define the problem, also
use one or more of the corresponding constraint classes included in
``parmoo.constraints.dtlz`` [NOT YET IMPLEMENTED].

The full list of simulation functions in this module includes the kernel
functions:
 * ``g1_sim``
 * ``g2_sim``
 * ``g3_sim``
 * ``g4_sim``

and the 9 DTLZ problems in simulation form, with each simulation output
corresponding to an objective:
 * ``dtlz1_sim``
 * ``dtlz2_sim``
 * ``dtlz3_sim``
 * ``dtlz4_sim``
 * ``dtlz5_sim``
 * ``dtlz6_sim``
 * ``dtlz7_sim``
 * ``dtlz8_sim``
 * ``dtlz9_sim``

"""

from parmoo.simulations import sim_func
from parmoo.util import to_array
import numpy as np


class g1_sim(sim_func):
    """ Class defining 1 of 4 kernel functions used in the DTLZ problem suite.

    g1 = 100 ( (n - o + 1) +
               sum_{i=o}^n ((x_i - offset)^2 - cos(20pi(x_i - offset))) )

    Contains 2 methods:
     * ``__init__(des, num_obj)``
     * ``__call__(x)``

    The ``__init__`` method inherits from the sim_func ABC.

    The ``__call__`` method performs an evaluation of the g1 kernel.

    """

    def __init__(self, des, num_obj=3, offset=0.5):
        """ Constructor for g1 class.

        Args:
            des (np.dtype): The numpy.dtype of the design variables.

            num_obj (int, optional): The number of objectives, which is
                used to calculate the value of g1. Note that regardless of
                the number of objectives, the number of simulation
                outputs from g1 is always 1.

            offset (optional, float): The location of the global minimizers
                is x_i = offset for i = number of objectives, ..., number of
                design variables. The default value is offset = 0.5.

        """

        super().__init__(des)
        __check_optionals__(num_obj=num_obj, offset=offset)
        self.o = num_obj
        self.offset = offset
        return

    def __call__(self, x):
        """ Define simulation evaluation.

        Args:
            x (numpy.array): A numpy.ndarray (unnamed) or numpy structured
                array (named), containing the design point to evaluate.

        Returns:
            numpy.ndarray: The output of this simulation for the input x.

        """

        # Extract x into xx, if names are used
        xx = to_array(x, self.des_type)
        # Calculate output
        result = (1 + self.n - self.o +
                  np.sum((xx[self.o-1:self.n] - self.offset) ** 2 -
                         np.cos(20.0 * np.pi *
                                (xx[self.o-1:self.n] - self.offset)))) * 100.0
        return np.array([result])


class g2_sim(sim_func):
    """ Class defining 2 of 4 kernel functions used in the DTLZ problem suite.

    g2 = (x_o - offset)^2 + ... + (x_n - offset)^2

    Contains 2 methods:
     * ``__init__(des)``
     * ``__call__(x)``

    The ``__init__`` method inherits from the sim_func ABC.

    The ``__call__`` method performs an evaluation of the g2 problem.

    """

    def __init__(self, des, num_obj=3, offset=0.5):
        """ Constructor for g2 class.

        Args:
            des (np.dtype): The numpy.dtype of the design variables.

            num_obj (int, optional): The number of objectives, which is
                used to calculate the value of g2. Note that regardless of
                the number of objectives, the number of simulation
                outputs from g2 is always 1.

            offset (optional, float): The location of the global minimizers
                is x_i = offset for i = number of objectives, ..., number of
                design variables. The default value is offset = 0.5.

        """

        super().__init__(des)
        __check_optionals__(num_obj=num_obj, offset=offset)
        self.o = num_obj
        self.offset = offset
        return

    def __call__(self, x):
        """ Define simulation evaluation.

        Args:
            x (numpy.array): A numpy.ndarray (unnamed) or numpy structured
                array (named), containing the design point to evaluate.

        Returns:
            numpy.ndarray: The output of this simulation for the input x.

        """

        # Extract x into xx, if names are used
        xx = to_array(x, self.des_type)
        return np.array([np.sum((xx[self.o-1:self.n] - self.offset) ** 2)])


class g3_sim(sim_func):
    """ Class defining 3 of 4 kernel functions used in the DTLZ problem suite.

    g3 = |x_o - offset|^.1 + ... + |x_n - offset|^.1

    Contains 2 methods:
     * ``__init__(des)``
     * ``__call__(x)``

    The ``__init__`` method inherits from the sim_func ABC.

    The ``__call__`` method performs an evaluation of the g3 problem.

    """

    def __init__(self, des, num_obj=3, offset=0.0):
        """ Constructor for g3 class.

        Args:
            des (np.dtype): The numpy.dtype of the design variables.

            num_obj (int, optional): The number of objectives, which is
                used to calculate the value of g3. Note that regardless of
                the number of objectives, the number of simulation
                outputs from g3 is always 1.

            offset (optional, float): The location of the global minimizers
                is x_i = offset for i = number of objectives, ..., number of
                design variables. The default value is offset = 0.0.

        """

        super().__init__(des)
        __check_optionals__(num_obj=num_obj, offset=offset)
        self.o = num_obj
        self.offset = offset
        return

    def __call__(self, x):
        """ Define simulation evaluation.

        Args:
            x (numpy.array): A numpy.ndarray (unnamed) or numpy structured
                array (named), containing the design point to evaluate.

        Returns:
            numpy.ndarray: The output of this simulation for the input x.

        """

        # Extract x into xx, if names are used
        xx = to_array(x, self.des_type)
        return np.array([np.sum(np.abs(xx[self.o-1:self.n] - self.offset)
                                ** 0.1)])


class g4_sim(sim_func):
    """ Class defining 4 of 4 kernel functions used in the DTLZ problem suite.

    g4 = 1 + (9 * (|x_o - offset| + ... + |x_n - offset|) / (n + 1 - o))

    Contains 2 methods:
     * ``__init__(des)``
     * ``__call__(x)``

    The ``__init__`` method inherits from the sim_func ABC.

    The ``__call__`` method performs an evaluation of the g4 problem.

    """

    def __init__(self, des, num_obj=3, offset=0.0):
        """ Constructor for g4 class.

        Args:
            des (np.dtype): The numpy.dtype of the design variables.

            num_obj (int, optional): The number of objectives, which is
                used to calculate the value of g4. Note that regardless of
                the number of objectives, the number of simulation
                outputs from g4 is always 1.

            offset (optional, float): The location of the global minimizers
                is x_i = offset for i = number of objectives, ..., number of
                design variables. The default value is offset = 0.0.

        """

        super().__init__(des)
        __check_optionals__(num_obj=num_obj, offset=offset)
        self.o = num_obj
        self.offset = offset
        return

    def __call__(self, x):
        """ Define simulation evaluation.

        Args:
            x (numpy.array): A numpy.ndarray (unnamed) or numpy structured
                array (named), containing the design point to evaluate.

        Returns:
            numpy.ndarray: The output of this simulation for the input x.

        """

        # Extract x into xx, if names are used
        xx = to_array(x, self.des_type)
        return np.array([(9 * np.sum(np.abs(xx[self.o-1:self.n] - self.offset))
                          / float(self.n + 1 - self.o)) + 1.0])


class dtlz1_sim(sim_func):
    """ Class defining the DTLZ1 problem with offset minimizer.

    DTLZ1 has a linear Pareto front, with all nondominated points
    on the hyperplane F_1 + F_2 + ... + F_o = 0.5.
    DTLZ1 has 11^k - 1 "local" Pareto fronts where k = n - o + 1, and
    1 "global" Pareto front.

    Contains 2 methods:
     * ``__init__(des)``
     * ``__call__(x)``

    The ``__init__`` method inherits from the sim_func ABC.

    The ``__call__`` method performs an evaluation of the DTLZ1 problem.

    """

    def __init__(self, des, num_obj=3, offset=0.5):
        """ Constructor for DTLZ1 class.

        Args:
            des (np.dtype): The numpy.dtype of the design variables.

            num_obj (int, optional): The number of objectives, which is
                used as the number of simulation outputs.

            offset (optional, float): The location of the global minimizers
                is x_i = offset for i = number of objectives, ..., number of
                design variables. The default value is offset = 0.5.

        """

        super().__init__(des)
        __check_optionals__(num_obj=num_obj, offset=offset)
        self.o = num_obj
        self.offset = offset
        self.ker = g1_sim(self.des_type, self.o, self.offset)
        return

    def __call__(self, x):
        """ Define simulation evaluation.

        Args:
            x (numpy.array): A numpy.ndarray (unnamed) or numpy structured
                array (named), containing the design point to evaluate.

        Returns:
            numpy.ndarray: The output of this simulation for the input x.

        """

        # Extract x into xx, if names are used
        xx = to_array(x, self.des_type)
        # Initialize output array
        fx = np.zeros(self.o)
        fx[:] = (1.0 + self.ker(x)[0]) / 2.0
        # Calculate the output array
        for i in range(self.o):
            for j in range(self.o - 1 - i):
                fx[i] *= xx[j]
            if i > 0:
                fx[i] *= (1.0 - xx[self.o - 1 - i])
        return fx


class dtlz2_sim(sim_func):
    """ Class defining the DTLZ2 problem with offset minimizer.

    DTLZ2 has a concave Pareto front, given by the unit sphere in
    objective space, restricted to the positive orthant.
    DTLZ2 has no "local" Pareto fronts, besides the true Pareto front.

    Contains 2 methods:
     * ``__init__(des, sim)``
     * ``__call__(x)``

    The ``__init__`` method inherits from the sim_func ABC.

    The ``__call__`` method performs an evaluation of the DTLZ2 problem.

    """

    def __init__(self, des, num_obj=3, offset=0.5):
        """ Constructor for DTLZ2 class.

        Args:
            des (np.dtype): The numpy.dtype of the design variables.

            num_obj (int, optional): The number of objectives, which is
                used as the number of simulation outputs.

            offset (optional, float): The location of the global minimizers
                is x_i = offset for i = number of objectives, ..., number of
                design variables. The default value is offset = 0.5.

        """

        super().__init__(des)
        __check_optionals__(num_obj=num_obj, offset=offset)
        self.o = num_obj
        self.offset = offset
        self.ker = g2_sim(self.des_type, self.o, self.offset)
        return

    def __call__(self, x):
        """ Define simulation evaluation.

        Args:
            x (numpy.array): A numpy.ndarray (unnamed) or numpy structured
                array (named), containing the design point to evaluate.

        Returns:
            numpy.ndarray: The output of this simulation for the input x.

        """

        # Extract x into xx, if names are used
        xx = to_array(x, self.des_type)
        # Initialize output array
        fx = np.zeros(self.o)
        fx[:] = (1.0 + self.ker(x)[0])
        # Calculate the output array
        for i in range(self.o):
            for j in range(self.o - 1 - i):
                fx[i] *= np.cos(np.pi * xx[j] / 2)
            if i > 0:
                fx[i] *= np.sin(np.pi * xx[self.o - 1 - i] / 2)
        return fx


class dtlz3_sim(sim_func):
    """ Class defining the DTLZ3 problem with offset minimizer.

    DTLZ3 has a concave Pareto front, given by the unit sphere in
    objective space, restricted to the positive orthant.
    DTLZ3 has 3^k - 1 "local" Pareto fronts where k = n - o + 1, and
    1 "global" Pareto front.

    Contains 2 methods:
     * ``__init__(des, sim)``
     * ``__call__(x)``

    The ``__init__`` method inherits from the sim_func ABC.

    The ``__call__`` method performs an evaluation of the DTLZ3 problem.

    """

    def __init__(self, des, num_obj=3, offset=0.5):
        """ Constructor for DTLZ3 class.

        Args:
            des (np.dtype): The numpy.dtype of the design variables.

            num_obj (int, optional): The number of objectives, which is
                used as the number of simulation outputs.

            offset (optional, float): The location of the global minimizers
                is x_i = offset for i = number of objectives, ..., number of
                design variables. The default value is offset = 0.5.

        """

        super().__init__(des)
        __check_optionals__(num_obj=num_obj, offset=offset)
        self.o = num_obj
        self.offset = offset
        self.ker = g1_sim(self.des_type, self.o, self.offset)
        return

    def __call__(self, x):
        """ Define simulation evaluation.

        Args:
            x (numpy.array): A numpy.ndarray (unnamed) or numpy structured
                array (named), containing the design point to evaluate.

        Returns:
            numpy.ndarray: The output of this simulation for the input x.

        """

        # Extract x into xx, if names are used
        xx = to_array(x, self.des_type)
        # Initialize output array
        fx = np.zeros(self.o)
        fx[:] = (1.0 + self.ker(x)[0])
        # Calculate the output array
        for i in range(self.o):
            for j in range(self.o - 1 - i):
                fx[i] *= np.cos(np.pi * xx[j] / 2)
            if i > 0:
                fx[i] *= np.sin(np.pi * xx[self.o - 1 - i] / 2)
        return fx


class dtlz4_sim(sim_func):
    """ Class defining the DTLZ4 problem with offset minimizer.

    DTLZ4 has a concave Pareto front, given by the unit sphere in
    objective space, restricted to the positive orthant.
    DTLZ4 has no "local" Pareto fronts, besides the true Pareto front,
    but by tuning the optional parameter alpha, one can adjust the
    solution density, making it harder for MOO algorithms to produce
    a uniform distribution of solutions.

    Contains 2 methods:
     * ``__init__(des, sim)``
     * ``__call__(x)``

    The ``__init__`` method inherits from the sim_func ABC.

    The ``__call__`` method performs an evaluation of the DTLZ4 problem.

    """

    def __init__(self, des, num_obj=3, offset=0.5, alpha=100.0):
        """ Constructor for DTLZ4 class.

        Args:
            des (np.dtype): The numpy.dtype of the design variables.

            num_obj (int, optional): The number of objectives, which is
                used as the number of simulation outputs.

            offset (optional, float): The location of the global minimizers
                is x_i = offset for i = number of objectives, ..., number of
                design variables. The default value is offset = 0.5.

            alpha (optional, float or int): The uniformity parameter used for
                controlling the uniformity of the distribution of solutions
                across the Pareto front. Must be greater than or equal to 1.
                A value of 1 results in DTLZ2. Default value is 100.0.

        """

        super().__init__(des)
        __check_optionals__(num_obj=num_obj, offset=offset, alpha=alpha)
        self.o = num_obj
        self.offset = offset
        self.alpha = alpha
        self.ker = g2_sim(self.des_type, self.o, self.offset)
        return

    def __call__(self, x):
        """ Define simulation evaluation.

        Args:
            x (numpy.array): A numpy.ndarray (unnamed) or numpy structured
                array (named), containing the design point to evaluate.

        Returns:
            numpy.ndarray: The output of this simulation for the input x.

        """

        # Extract x into xx, if names are used
        xx = to_array(x, self.des_type)
        # Initialize output array
        fx = np.zeros(self.o)
        fx[:] = (1.0 + self.ker(x)[0])
        # Calculate the output array
        for i in range(self.o):
            for j in range(self.o - 1 - i):
                fx[i] *= np.cos(np.pi * xx[j] ** self.alpha / 2)
            if i > 0:
                fx[i] *= np.sin(np.pi * xx[self.o - 1 - i] ** self.alpha / 2)
        return fx


class dtlz5_sim(sim_func):
    """ Class defining the DTLZ5 problem with offset minimizer.

    DTLZ5 has a lower-dimensional Pareto front embedded in the objective
    space, given by an arc of the unit sphere in the positive orthant.
    DTLZ5 has no "local" Pareto fronts, besides the true Pareto front.

    Contains 2 methods:
     * ``__init__(des, sim)``
     * ``__call__(x)``

    The ``__init__`` method inherits from the sim_func ABC.

    The ``__call__`` method performs an evaluation of the DTLZ5 problem.

    """

    def __init__(self, des, num_obj=3, offset=0.5):
        """ Constructor for DTLZ5 class.

        Args:
            des (np.dtype): The numpy.dtype of the design variables.

            num_obj (int, optional): The number of objectives, which is
                used as the number of simulation outputs.

            offset (optional, float): The location of the global minimizers
                is x_i = offset for i = number of objectives, ..., number of
                design variables. The default value is offset = 0.5.

        """

        super().__init__(des)
        __check_optionals__(num_obj=num_obj, offset=offset)
        self.o = num_obj
        self.offset = offset
        self.ker = g2_sim(self.des_type, self.o, self.offset)
        return

    def __call__(self, x):
        """ Define simulation evaluation.

        Args:
            x (numpy.array): A numpy.ndarray (unnamed) or numpy structured
                array (named), containing the design point to evaluate.

        Returns:
            numpy.ndarray: The output of this simulation for the input x.

        """

        # Extract x into xx, if names are used
        xx = to_array(x, self.des_type)
        # Calculate theta values
        theta = np.zeros(self.o)
        g2x = self.ker(x)[0]
        theta[0] = xx[0]
        for i in range(1, self.o):
            theta[i] = (1 + 2 * g2x * xx[i]) / (2 * (1 + g2x))
        # Initialize output array
        fx = np.zeros(self.o)
        fx[:] = (1.0 + g2x)
        # Calculate the output array
        for i in range(self.o):
            for j in range(self.o - 1 - i):
                fx[i] *= np.cos(np.pi * theta[j] / 2)
            if i > 0:
                fx[i] *= np.sin(np.pi * theta[self.o - 1 - i] / 2)
        return fx


class dtlz6_sim(sim_func):
    """ Class defining the DTLZ6 problem with offset minimizer.

    DTLZ6 has a lower-dimensional Pareto front embedded in the objective
    space, given by an arc of the unit sphere in the positive orthant.
    DTLZ6 has no "local" Pareto fronts, but tends to show very little
    improvement until the algorithm is very close to its solution set.

    Contains 2 methods:
     * ``__init__(des, sim)``
     * ``__call__(x)``

    The ``__init__`` method inherits from the sim_func ABC.

    The ``__call__`` method performs an evaluation of the DTLZ6 problem.

    """

    def __init__(self, des, num_obj=3, offset=0.0):
        """ Constructor for DTLZ6 class.

        Args:
            des (np.dtype): The numpy.dtype of the design variables.

            num_obj (int, optional): The number of objectives, which is
                used as the number of simulation outputs.

            offset (optional, float): The location of the global minimizers
                is x_i = offset for i = number of objectives, ..., number of
                design variables. The default value is offset = 0.0.

        """

        super().__init__(des)
        __check_optionals__(num_obj=num_obj, offset=offset)
        self.o = num_obj
        self.offset = offset
        self.ker = g3_sim(self.des_type, self.o, self.offset)
        return

    def __call__(self, x):
        """ Define simulation evaluation.

        Args:
            x (numpy.array): A numpy.ndarray (unnamed) or numpy structured
                array (named), containing the design point to evaluate.

        Returns:
            numpy.ndarray: The output of this simulation for the input x.

        """

        # Extract x into xx, if names are used
        xx = to_array(x, self.des_type)
        # Calculate theta values
        theta = np.zeros(self.o)
        g3x = self.ker(x)[0]
        theta[0] = xx[0]
        for i in range(1, self.o):
            theta[i] = (1 + 2 * g3x * xx[i]) / (2 * (1 + g3x))
        # Initialize output array
        fx = np.zeros(self.o)
        fx[:] = (1.0 + g3x)
        # Calculate the output array
        for i in range(self.o):
            for j in range(self.o - 1 - i):
                fx[i] *= np.cos(np.pi * theta[j] / 2)
            if i > 0:
                fx[i] *= np.sin(np.pi * theta[self.o - 1 - i] / 2)
        return fx


class dtlz7_sim(sim_func):
    """ Class defining the DTLZ7 problem with offset minimizer.

    DTLZ7 has a discontinuous Pareto front, with solutions on the
    2^(o-1) discontinuous nondominated regions of the surface:

    F_m = o - F_1 (1 + sin(3pi F_1)) - ... - F_{o-1} (1 + sin3pi F_{o-1}).

    Contains 2 methods:
     * ``__init__(des, sim)``
     * ``__call__(x)``

    The ``__init__`` method inherits from the sim_func ABC.

    The ``__call__`` method performs an evaluation of the DTLZ7 problem.

    """

    def __init__(self, des, num_obj=3, offset=0.0):
        """ Constructor for DTLZ7 class.

        Args:
            des (np.dtype): The numpy.dtype of the design variables.

            num_obj (int, optional): The number of objectives, which is
                used as the number of simulation outputs.

            offset (optional, float): The location of the global minimizers
                is x_i = offset for i = number of objectives, ..., number of
                design variables. The default value is offset = 0.0.

        """

        super().__init__(des)
        __check_optionals__(num_obj=num_obj, offset=offset)
        self.o = num_obj
        self.offset = offset
        self.ker = g4_sim(self.des_type, self.o, self.offset)
        return

    def __call__(self, x):
        """ Define simulation evaluation.

        Args:
            x (numpy.array): A numpy.ndarray (unnamed) or numpy structured
                array (named), containing the design point to evaluate.

        Returns:
            numpy.ndarray: The output of this simulation for the input x.

        """

        # Extract x into xx, if names are used
        xx = to_array(x, self.des_type)
        # Initialize first o-1 entries in the output array
        fx = np.zeros(self.o)
        fx[:self.o-1] = xx[:self.o-1]
        # Calculate kernel functions
        gx = 1.0 + self.ker(x)[0]
        hx = (-np.sum(xx[:self.o-1] *
                      (1.0 + np.sin(3.0 * np.pi * xx[:self.o-1])) / gx)
              + float(self.o))
        # Calculate the last entry in the output array
        fx[self.o-1] = gx * hx
        return fx


class dtlz8_sim(sim_func):
    """ Class defining the DTLZ8 problem with offset minimizer.

    DTLZ8 is a constrained MOOP, whose Pareto front combines a region
    of a plane with a line segment. To fully define DTLZ8, you must also
    use the parmoo.constraints.dtlz module.

    Contains 2 methods:
     * ``__init__(des, sim)``
     * ``__call__(x)``

    The ``__init__`` method inherits from the sim_func ABC.

    The ``__call__`` method performs an evaluation of the DTLZ8 problem.

    """

    def __init__(self, des, num_obj=3, offset=0.0):
        """ Constructor for DTLZ8 class.

        Args:
            des (np.dtype): The numpy.dtype of the design variables.

            num_obj (int, optional): The number of objectives, which is
                used as the number of simulation outputs.

            offset (optional, float): The location of the global minimizers
                is x_i = offset for i = number of objectives, ..., number of
                design variables. The default value is offset = 0.0.

        """

        super().__init__(des)
        __check_optionals__(num_obj=num_obj, offset=offset)
        self.o = num_obj
        self.offset = offset
        return

    def __call__(self, x):
        """ Define simulation evaluation.

        Args:
            x (numpy.array): A numpy.ndarray (unnamed) or numpy structured
                array (named), containing the design point to evaluate.

        Returns:
            numpy.ndarray: The output of this simulation for the input x.

        """

        # Extract x into xx, if names are used
        xx = to_array(x, self.des_type)
        # Initialize the output array
        fx = np.zeros(self.o)
        # Calculate outputs
        for i in range(self.o):
            start = i * self.n // self.o
            stop = (i + 1) * self.n // self.o
            fx[i] = np.sum(np.abs(xx[start:stop] - self.offset))
        return fx


class dtlz9_sim(sim_func):
    """ Class defining the DTLZ9 problem with offset minimizer.

    DTLZ9 is a constrained MOOP, whose Pareto front is a subregion of
    the arc traced out by the solution to DTLZ5. To fully define DTLZ8,
    you must also use the parmoo.constraints.dtlz module.

    Contains 2 methods:
     * ``__init__(des, sim)``
     * ``__call__(x)``

    The ``__init__`` method inherits from the sim_func ABC.

    The ``__call__`` method performs an evaluation of the DTLZ9 problem.

    """

    def __init__(self, des, num_obj=3, offset=0.0):
        """ Constructor for DTLZ9 class.

        Args:
            des (np.dtype): The numpy.dtype of the design variables.

            num_obj (int, optional): The number of objectives, which is
                used as the number of simulation outputs.

            offset (optional, float): The location of the global minimizers
                is x_i = offset for i = number of objectives, ..., number of
                design variables. The default value is offset = 0.0.

        """

        super().__init__(des)
        __check_optionals__(num_obj=num_obj, offset=offset)
        self.o = num_obj
        self.offset = offset
        return

    def __call__(self, x):
        """ Define simulation evaluation.

        Args:
            x (numpy.array): A numpy.ndarray (unnamed) or numpy structured
                array (named), containing the design point to evaluate.

        Returns:
            numpy.ndarray: The output of this simulation for the input x.

        """

        # Extract x into xx, if names are used
        xx = to_array(x, self.des_type)
        # Initialize the output array
        fx = np.zeros(self.o)
        # Calculate outputs
        for i in range(self.o):
            start = i * self.n // self.o
            stop = (i + 1) * self.n // self.o
            fx[i] = np.sum(np.abs(xx[start:stop] - self.offset) ** 0.1)
        return fx


def __check_optionals__(num_obj=3, offset=0.5, alpha=100.0):
    """ Check DTLZ optional inputs for illegal values.

    Not recommended for external usage.

    """

    if not isinstance(offset, float):
        raise TypeError("optional input offset must have the float type")
    if offset < 0 or offset > 1:
        raise ValueError("offset must be in the range [0, 1]")
    if not isinstance(num_obj, int):
        raise TypeError("optional input num_obj must be an int type")
    if num_obj < 1:
        raise ValueError("optional input num_obj must be greater than 0")
    if not (isinstance(alpha, int) or isinstance(alpha, float)):
        raise TypeError("alpha must be a numeric type")
    if alpha < 1:
        raise ValueError("alpha must be at least 1")
    return




def test_Linear():
    """ Test the Linear model class in surrogates.py.

    Generate random x and y values for a 3D input space and 2D output space,
    then use the Linear class to fit linear models to the data.

    Check that the linear models interpolate the original data.

    """

    from jax import config, jacrev
    config.update("jax_enable_x64", True)
    import numpy as np
    import os
    from parmoo.surrogates import Linear
    import pytest

    # Try some bad initializations to test error handling
    with pytest.raises(ValueError):
        Linear(2, np.zeros(3), np.ones(3), {'des_tols': np.zeros(3)})
    with pytest.raises(ValueError):
        Linear(2, np.zeros(3), np.ones(3), {'des_tols': np.zeros(2)})
    with pytest.raises(ValueError):
        Linear(2, np.zeros(3), np.ones(3), {'des_tols': 0.1})
    # Create 2 identical Linear models
    lsm1 = Linear(2, np.zeros(3), np.ones(3), {})
    lsm2 = Linear(2, np.zeros(3), np.ones(3), {})
    # Generate some random data with 3 design variables and 2 outputs
    x_vals1 = np.random.random_sample((3, 3))
    y_vals1 = np.random.random_sample((3, 2))
    x_vals2 = np.random.random_sample((3, 3))
    y_vals2 = np.random.random_sample((3, 2))
    x_vals_full = np.concatenate((x_vals1, x_vals2), axis=0)
    y_vals_full = np.concatenate((y_vals1, y_vals2), axis=0)
    # Try to fit/update with illegal data to test error handling
    with pytest.raises(TypeError):
        lsm1.fit(0, y_vals1)
    with pytest.raises(TypeError):
        lsm1.fit(x_vals1, 0)
    with pytest.raises(ValueError):
        lsm1.fit(np.zeros((0, 3)), np.zeros((0, 2)))
    with pytest.raises(ValueError):
        lsm1.fit(np.zeros((10, 3)), np.zeros((10, 3)))
    with pytest.raises(ValueError):
        lsm1.fit(np.zeros((10, 3)), np.zeros((9, 2)))
    with pytest.raises(TypeError):
        lsm1.update(0, y_vals1)
    with pytest.raises(TypeError):
        lsm1.update(x_vals1, 0)
    with pytest.raises(ValueError):
        lsm1.update(np.zeros((10, 3)), np.zeros((10, 3)))
    with pytest.raises(ValueError):
        lsm1.update(np.zeros((10, 3)), np.zeros((9, 2)))
    # Fit both models, with and without using the update method
    lsm1.fit(x_vals1, y_vals1)
    lsm1.update(x_vals2, y_vals2)
    lsm1.update(np.zeros((0, 3)), np.zeros((0, 2)))    # Update with no data
    lsm1.setTrustRegion(0.5 * np.ones(3), np.ones(3) * 0.5)
    lsm2.fit(x_vals_full, y_vals_full)
    lsm2.setTrustRegion(0.5 * np.ones(3), np.ones(3) * 0.5)
    # Try to set the center with illegal values to test error handling
    with pytest.raises(TypeError):
        lsm1.setTrustRegion(5, np.ones(3))
    with pytest.raises(ValueError):
        lsm1.setTrustRegion(np.zeros(5), np.ones(3))
    with pytest.raises(ValueError):
        lsm1.setTrustRegion(-np.ones(3), np.ones(3))
    # Try to set the radius with illegal values to test error handling
    with pytest.raises(TypeError):
        lsm1.setTrustRegion(np.ones(3), 5)
    with pytest.raises(ValueError):
        lsm1.setTrustRegion(np.ones(3), np.ones(5))
    with pytest.raises(ValueError):
        lsm1.setTrustRegion(np.ones(3), -np.ones(3))
    # Try a bad improvement call to test error handling
    with pytest.raises(TypeError):
        lsm1.improve(5, False)
    with pytest.raises(ValueError):
        lsm1.improve(np.zeros(2), False)
    with pytest.raises(ValueError):
        lsm1.improve(-np.ones(3), False)
    # Check that the models match on a random evaluation point
    x = np.random.random_sample((3))
    assert (all(lsm1.evaluate(x) == lsm2.evaluate(x)))
    # Check that the models interpolate, up to 8 decimal digits of precision
    for xi, yi in zip(x_vals_full, y_vals_full):
        lsm1.setTrustRegion(xi, np.ones(3) * 0.1)
        lsm2.setTrustRegion(xi, np.ones(3) * 0.1)
        assert (np.linalg.norm(lsm1.evaluate(xi) - yi) < 1.0e-8)
        assert (np.linalg.norm(lsm2.evaluate(xi) - yi) < 1.0e-8)
    # Check that the models compute the same grad, up to 8 digits of precision
    lsm1.setTrustRegion(0.5 * np.ones(3), np.ones(3) * 0.5)
    lsm2.setTrustRegion(0.5 * np.ones(3), np.ones(3) * 0.5)
    lsm1_grad = jacrev(lsm1.evaluate)
    lsm2_grad = jacrev(lsm2.evaluate)
    for i in range(x_vals_full.shape[0]):
        assert (np.linalg.norm(lsm1_grad(x_vals_full[i]) -
                               lsm2_grad(x_vals_full[i])) < 1.0e-8)
    # Check that the model generates feasible local improvement points
    x_vals3 = np.eye(3)
    x_vals3 = np.append(x_vals3, [[0.5, 0.5, 0.5]], axis=0)
    y_vals3 = np.asarray([[np.dot(xi, xi)] for xi in x_vals3])
    lsm3 = Linear(1, np.zeros(3), np.ones(3), {'tail_order': 0})
    lsm3.fit(x_vals3, y_vals3)
    lsm3.setTrustRegion(x_vals3[-1], np.ones(3) * 0.5)
    for i in range(4):
        x_improv = lsm3.improve(np.zeros(3), False)
        assert (np.all(x_improv[0] <= np.ones(3)) and
                np.all(x_improv[0] >= np.zeros(3)))
    # Check that the model generates feasible global improvement points
    x_improv = lsm3.improve(x_vals3[-1], True)
    assert (np.all(x_improv[0] <= np.ones(3)) and
            np.all(x_improv[0] >= np.zeros(3)))
    # Check that model generates good improvements when points are bunched
    x_new = np.ones((3, 3))
    x_new = x_new * 0.5
    f_new = np.zeros((3, 1))
    for i in range(3):
        x_new[i, i] = 0.50000001
        f_new[i, 0] = np.dot(x_new[i, :], x_new[i, :])
    lsm3.update(x_new, f_new)
    lsm3.setTrustRegion(0.5 * np.ones(3), np.ones(3) * 1.0e-4)
    x_improv = lsm3.improve(np.asarray([0.5, 0.5, 0.5]), False)
    assert (np.all(x_improv[0] <= np.ones(3)) and
            np.all(x_improv[0] >= np.zeros(3)))
    assert (np.all(np.linalg.norm(x_improv[0] - np.asarray([0.5, 0.5, 0.5]))
            > 0.00000001))
    # Now create a really tiny design space with a large tolerance
    lsm4 = Linear(1, np.zeros(1), np.ones(1), {'des_tols': 0.3 * np.ones(1)})
    xdat4 = np.asarray([[0.4], [0.6]])
    ydat4 = np.asarray([[0.4], [0.6]])
    lsm4.fit(xdat4, ydat4)
    # Test that improve() is able to find points outside the design tolerance
    for i in range(5):
        x_improv = lsm4.improve(xdat4[0], False)
        assert (x_improv[0][0] < 0.1 or x_improv[0][0] > 0.9)
    for i in range(5):
        x_improv = lsm4.improve(xdat4[0], True)
        assert (x_improv[0][0] < 0.1 or x_improv[0][0] > 0.9)
    # Now fit datapoints in a plane
    x_vals3 = np.zeros((4, 3))
    x_vals3[1, 0] = 0.1
    x_vals3[2, 0] = 0.2
    x_vals3[3, 0] = 0.3
    y_vals3 = np.ones((4, 2))
    lsm5 = Linear(2, np.zeros(3), np.ones(3), {})
    lsm5.fit(x_vals3, y_vals3)
    lsm5.update(x_vals3, y_vals3)
    lsm5.setTrustRegion(x_vals3[-1], np.ones(3) * 0.2)
    # Test save and load
    lsm5.save("parmoo.surrogate")
    lsm4.load("parmoo.surrogate")
    xx = np.random.random_sample(3)
    assert (np.all(lsm4.evaluate(xx) == lsm5.evaluate(xx)))
    os.remove("parmoo.surrogate")
    # Generate a simple 1D model and check it is accurate
    x_vals4 = np.array([[0], [1]])
    y_vals4 = np.array([[1], [1]])
    lsm6 = Linear(1, np.zeros(1), np.ones(1), {})
    lsm6.fit(x_vals4, y_vals4)
    lsm6.setTrustRegion(np.array([0.5]), np.ones(1) * 0.1)
    assert (np.linalg.norm(lsm6.evaluate(np.array([0.5])) - 1.0) < 1.0e-8)
    assert (np.linalg.norm(jacrev(lsm6.evaluate)(np.array([0.5]))) < 1.0e-8)


if __name__ == "__main__":
    test_Linear()




""" Abstract base classes (ABCs) for ParMOO project.

This module contains several ABCs that can be used to extend and customize
ParMOO's solver components and/or supported problem types.

The classes include:
 * AcquisitionFunction
 * CompositeFunction
 * Embedder
 * GlobalSearch
 * SurrogateFunction
 * SurrogateOptimizer

"""

from abc import ABC, abstractmethod
import inspect
import numpy as np
from scipy.stats import tstd


class AcquisitionFunction(ABC):
    """ ABC describing acquisition functions.

    This class contains the following methods:
     * ``setTarget(data, penalty_func)``
     * ``scalarize(f_vals, x_vals, s_vals_mean, s_vals_sd)``
     * ``useSD()``
     * ``save(filename)``
     * ``load(filename)``

    """

    @abstractmethod
    def __init__(self, o, lb, ub, hyperparams):
        """ Constructor for the AcquisitionFunction class.

        Args:
            o (int): The number of objectives.

            lb (ndarray): A 1D array of lower bounds for the design space.

            ub (ndarray): A 1D array of upper bounds for the design space.

            hyperparams (dict): A dictionary of hyperparameters that are
                passed to the acquisition function.

        Returns:
            AcquisitionFunction: A new AcquisitionFunction object.

        """

    @abstractmethod
    def setTarget(self, data, penalty_func):
        """ Set a new target value or region for the AcquisitionFunction.

        Args:
            data (dict): A dictionary specifying the current function
                evaluation database. It contains two mandatory fields:
                 * 'x_vals' (ndarray): A 2D array of design points.
                 * 'f_vals' (ndarray): A 2D array of corresponding objectives.

            penalty_func (function): A function of one (x) or two (x, sx)
                inputs that evaluates all (penalized) objective scores.

        Returns:
            ndarray: A 1D array containing a feasible starting point
            for the scalarized problem.

        """

    def useSD(self):
        """ Query whether this method uses uncertainties.

        When False, allows users to shortcut expensive uncertainty
        computations.

        Default implementation returns True, requiring full uncertainty
        computation for applicable models.

        """

        return True

    @abstractmethod
    def scalarize(self, f_vals, x_vals, s_vals_mean, s_vals_sd):
        """ Scalarize a vector-valued function using the AcquisitionFunction.

        Note: For best performance, make sure that jax can jit this method.

        Additionally, for compatibility with gradient-based solvers,
        this method must be implemented in jax and be differentiable
        via the jax.jacrev() tool.

        Args:
            f_vals (ndarray): A 1D array specifying a vector of function
                values to be scalarized.

            x_vals (ndarray): A 1D array specifying a vector the design
                point corresponding to f_vals.

            s_vals_mean (ndarray): A 1D array specifying the expected value
                of the simulation outputs for the x value being scalarized.

            s_vals_sd (ndarray): A 1D array specifying the standard deviation
                for each of the simulation outputs.

        Returns:
            float: The scalarized value.

        """

    def save(self, filename):
        """ Save important data from this class so that it can be reloaded.

        Note: If this function is left unimplemented, ParMOO will reinitialize
        a fresh instance after a save/load. If this is the desired behavior,
        then this method and the load method need not be implemented.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data should be saved.

        """

        raise NotImplementedError("This class method has not been implemented")

    def load(self, filename):
        """ Reload important data into this class after a previous save.

        Note: If this function is left unimplemented, ParMOO will reinitialize
        a fresh instance after a save/load. If this is the desired behavior,
        then this method and the save method need not be implemented.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data has been saved.

        """

        raise NotImplementedError("This class method has not been implemented")


class CompositeFunction(ABC):
    """ ABC defining ParMOO composite functions.

    Extend this class to create a callable object that matches ParMOO can
    use as a composite function, such as:
     - an Objective,
     - a Constraint, or
     - a Gradient.

    Contains 2 methods:
     * ``__init__(des_type, sim_type)``
     * ``__call__(x, sx)``

    The ``__init__`` method is already implemented, and is the constructor.
    It can be overwritten if additional inputs (besides the design variable
    and simulation output types) are needed.

    The ``__call__`` method is left to be implemented, and performs the
    composite function evaluation.

    """

    __slots__ = ['n', 'm', 'des_type', 'sim_type']

    def __init__(self, des_type, sim_type):
        """ Constructor for CompositeFunction class.

        Args:
            des_type (np.dtype): The numpy.dtype of the design variables.

            sim_type (np.dtype): The numpy.dtype of the simulation outputs.

        """

        # Try to read design variable type
        try:
            self.des_type = np.dtype(des_type)
        except TypeError:
            raise TypeError("des_type must contain a valid numpy.dtype")
        self.n = len(self.des_type.names)
        if self.n <= 0:
            raise ValueError("An illegal des_type was given")
        # Try to read simulation variable type
        try:
            self.sim_type = np.dtype(sim_type)
        except TypeError:
            raise TypeError("sim_type must contain a valid numpy.dtype")
        self.m = 0
        for name in self.sim_type.names:
            self.m += np.maximum(np.sum(self.sim_type[name].shape), 1)
        if self.m <= 0:
            raise ValueError("An illegal sim_type was given")
        return

    @abstractmethod
    def __call__(self, x, sx):
        """ Make CompositeFunction objects callable.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            float: The output of this objective for the input x (when
            defining objectives and constraint functions).

            OR

            dict, dict: Dictionaries with the same keys as x and sx, whose
            corresponding values contain the partials with respect to x and
            sx, respectively.

        """


class Embedder(ABC):
    """ ABC describing the embedding of design variables.

    This class contains the following methods:
     * ``getLatentDesTols()``
     * ``getFeatureDesTols()``
     * ``getEmbeddingSize()``
     * ``getInputType()``
     * ``getLowerBounds()``
     * ``getUpperBounds()``
     * ``embed(x)``
     * ``embed_grad(dx)``
     * ``extract(x)``

    """

    @abstractmethod
    def __init__(self, settings):
        """ Constructor for the Embedder class.

        Args:
            settings (dict): Contains any variable information that the user
                might need to provide.

        Returns:
            Embedder: A new Embedder object.

        """

    @abstractmethod
    def getLatentDesTols(self):
        """ Get the design tolerances along each dimension of the embedding.

        Returns:
            numpy.ndarray: array of design space tolerances after embedding.

        """

    @abstractmethod
    def getFeatureDesTols(self):
        """ Get the design tolerances in the feature space (pre-embedding).

        Returns:
            float: the design tolerance in the feature space -- a value of
            0 indicates that this is a discrete variable.

        """

    @abstractmethod
    def getEmbeddingSize(self):
        """ Get the dimension of the latent (embedded) space.

        Returns:
            int: the dimension of the latent space produced.

        """

    @abstractmethod
    def getInputType(self):
        """ Get the input type for this embedder.

        Note: Whatever the input type, the output type must always be a
        ndarray of one or more continuous variables in some range [lb, ub].

        Returns:
            str: A numpy string representation of the input type from the
            feature space.
            Currently supported values are: ["f8", "i4", "a25", or "u25"].

        """

    @abstractmethod
    def getLowerBounds(self):
        """ Get a vector of lower bounds for the embedded (latent) space.

        Returns:
            ndarray: A 1D array of lower bounds in embedded space whose size
                matches the output of ``getEmbeddingSize()``.

        """

    @abstractmethod
    def getUpperBounds(self):
        """ Get a vector of upper bounds for the embedded (latent) space.

        Returns:
            ndarray: A 1D array of upper bounds in embedded space whose size
                matches the output of ``getEmbeddingSize()``.

        """

    def embed(self, x):
        """ Embed a design input as an n-dimensional vector for ParMOO.

        Note: For best performance, make sure that jax can jit this method.

        Args:
            x (stype): The value of the design variable to embed, where
                stype matches the numpy-string type specified by
                getInputType().

        Returns:
            ndarray: A 1D array whose size matches the output of
            getEmbeddingSize() containing the embedding of x.

        """

        raise NotImplementedError("This Embedder has not implemented an "
                                  "embed method yet.")

    def embed_grad(self, dx):
        """ Embed a partial design gradient as a vector for ParMOO.

        Note: If not implemented, ParMOO will still work with gradient-free
        methods, but will not support autograd features.

        For best performance, make sure that jax can jit this method.

        Args:
            dx (float): The partial design gradient to embed.

        Returns:
            numpy.ndarray: A numpy array of length 1 containing a
            rescaling of x

        """

        raise NotImplementedError("This Embedder has not implemented an "
                                  "embed_grad method yet.")

    def extract(self, x):
        """ Extract a design input from an n-dimensional vector for ParMOO.

        Note: For best performance, make sure that jax can jit this method.

        Args:
            x (ndarray): A 1D array whose size matches the output of
                getEmbeddingSize() containing the embedding of x.

        Returns:
            stype: The value of the design variable to embed, where stype
            matches the numpy-string type specified by getInputType().

        """

        raise NotImplementedError("This Embedder has not implemented an "
                                  "extract method yet.")


class GlobalSearch(ABC):
    """ ABC describing global search techniques.

    This class contains the following methods.
     * ``startSearch(lb, ub)``
     * ``resumeSearch()``
     * ``save(filename)``
     * ``load(filename)``

    """

    @abstractmethod
    def __init__(self, o, lb, ub, hyperparams):
        """ Constructor for the GlobalSearch class.

        Args:
            o (int): The number of objectives.

            lb (ndarray): A 1D array of lower bounds for the design space.

            ub (ndarray): A 1D array of upper bounds for the design space.

            hyperparams (dict): A dictionary of hyperparameters for the
                global search. It may contain any inputs specific to the
                search algorithm.

        Returns:
            GlobalSearch: A new GlobalSearch object.

        """

    @abstractmethod
    def startSearch(self, lb, ub):
        """ Begin a new global search.

        Args:
            lb (ndarray): A 1D array of lower bounds for the design space.

            ub (ndarray): A 1D array of upper bounds for the design space.

        Returns:
            ndarray: A 2D design matrix.

        """

    def resumeSearch(self):
        """ Resume a global search.

        Returns:
            ndarray: A 2D design matrix.

        """

        raise NotImplementedError("This class method has not been implemented")

    def save(self, filename):
        """ Save important data from this class so that it can be reloaded.

        Note: If this function is left unimplemented, ParMOO will reinitialize
        a fresh instance after a save/load. If this is the desired behavior,
        then this method and the load method need not be implemented.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data should be saved.

        """

        raise NotImplementedError("This class method has not been implemented")

    def load(self, filename):
        """ Reload important data into this class after a previous save.

        Note: If this function is left unimplemented, ParMOO will reinitialize
        a fresh instance after a save/load. If this is the desired behavior,
        then this method and the save method need not be implemented.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data has been saved.

        """

        raise NotImplementedError("This class method has not been implemented")


class SurrogateFunction(ABC):
    """ ABC describing surrogate functions.

    This class contains the following methods.
     * ``fit(x, f)``
     * ``update(x, f)``
     * ``setTrustRegion(center, radius)`` (default implementation provided)
     * ``evaluate(x)``
     * ``stdDev(x)``
     * ``improve(x, global_improv)`` (default implementation provided)
     * ``save(filename)``
     * ``load(filename)``

    """

    @abstractmethod
    def __init__(self, m, lb, ub, hyperparams):
        """ Constructor for the SurrogateFunction class.

        Args:
            m (int): The number of objectives to fit.

            lb (ndarray): A 1D array of lower bounds for the design space.

            ub (ndarray): A 1D array of upper bounds for the design space.

            hyperparams (dict): A dictionary of hyperparameters to be used
                by the surrogate models, including:
                 * des_tols (ndarray, optional): A 1D array whose length
                   matches lb and ub. Each entry is a number (greater than 0)
                   specifying the design space tolerance for that variable.

        Returns:
            SurrogateFunction: A new SurrogateFunction object.

        """

    @abstractmethod
    def fit(self, x, f):
        """ Fit a new surrogate to the given data.

        Args:
             x (ndarray): A 2D array containing the design points to fit.

             f (ndarray): A 2D array of the corresponding objectives values.

        """

    @abstractmethod
    def update(self, x, f):
        """ Update an existing surrogate model using new data.

        Args:
             x (ndarray): A 2D array containing new design points to fit.

             f (ndarray): A 2D array of the corresponding objectives values.

        """

    def setTrustRegion(self, center, radius):
        """ Alert the surrogate of the trust-region center and radius.

        Default implementation does nothing, which would be the case for a
        global surrogate model.

        Args:
            center (ndarray): A 1D array containing the center for a local fit.

            radius (ndarray or float): The radius for the local fit.

        """

        return

    @abstractmethod
    def evaluate(self, x):
        """ Evaluate the surrogate at a design point.

        Note: For best performance, make sure that jax can jit this method.

        Additionally, for compatibility with gradient-based solvers,
        this method must be implemented in jax and be differentiable
        via the jax.jacrev() tool.

        Args:
            x (ndarray): A 1D array containing the design point to evaluate.

        Returns:
            ndarray: A 1D array containing the predicted outputs at x.

        """

    def stdDev(self, x):
        """ Evaluate the standard deviation of the surrogate at x.

        Note: this method need not be implemented when the acquisition
        function does not use the model uncertainty.

        Additionally, for compatibility with gradient-based solvers,
        this method must be implemented in jax and be differentiable
        via the jax.jacrev() tool.

        Args:
            x (ndarray): A 1D array containing the design point to evaluate.

        Returns:
            ndarray: A 1D array containing the output standard deviation at x.

        """

        raise NotImplementedError("This class method has not been implemented")

    def improve(self, x, global_improv):
        """ Suggests a design to evaluate to improve the surrogate near x.

        A default implementation is given based on random sampling.
        Re-implement the improve method to overwrite the default
        policy.

        Args:
            x (ndarray): A 1D array containing a design point where greater
                accuracy is needed.

            global_improv (Boolean): When True, ignore the value of x and
                seek global model improvement.

        Returns:
            ndarray: A 2D array containing a list of (at least 1) design points
            that could be evaluated to improve the surrogate model's accuracy.

        """

        # Check that the x is legal
        try:
            if x.size != self.n:
                raise ValueError("x must have length n")
            elif (np.any(x < self.lb - self.eps) or
                  np.any(x > self.ub + self.eps)):
                raise ValueError("x cannot be infeasible")
        except AttributeError:
            raise TypeError("x must be a numpy array-like object")
        # Allocate the output array.
        x_new = np.zeros(self.n)
        if global_improv:
            # If global improvement has been specified, randomly select a
            # point from within the bound constraints.
            x_new[:] = self.lb[:] + (np.random.random(self.n)
                                     * (self.ub[:] - self.lb[:]))
            while any([np.all(np.abs(x_new - xj) < self.eps)
                       for xj in self.x_vals]):
                x_new[:] = self.lb[:] + (np.random.random(self.n)
                                         * (self.ub[:] - self.lb[:]))
        else:
            # Find the n+1 closest points to x in the current database
            diffs = np.asarray([np.abs(x - xj) / self.eps
                                for xj in self.x_vals])
            dists = np.asarray([np.amax(dj) for dj in diffs])
            inds = np.argsort(dists)
            diffs = diffs[inds]
            if dists[inds[self.n]] > 1.5:
                # Calculate the normalized sample standard dev along each axis
                stddev = np.asarray(tstd(diffs[:self.n+1], axis=0))
                stddev[:] = np.maximum(stddev, np.ones(self.n))
                stddev[:] = stddev[:] / np.amin(stddev)
                # Sample within B(x, dists[inds[self.n]] / stddev)
                rad = (dists[inds[self.n]] * self.eps) / stddev
                x_new = np.fmin(np.fmax(2.0 * (np.random.random(self.n) - 0.5)
                                        * rad[:] + x, self.lb), self.ub)
                while any([np.all(np.abs(x_new - xj) < self.eps)
                           for xj in self.x_vals]):
                    x_new = np.fmin(np.fmax(2.0 *
                                            (np.random.random(self.n) - 0.5)
                                            * rad[:] + x, self.lb), self.ub)
            else:
                # If the n+1st nearest point is too close, use global_improv.
                x_new[:] = self.lb[:] + np.random.random(self.n) \
                           * (self.ub[:] - self.lb[:])
                # If the nearest point is too close, resample.
                while any([np.all(np.abs(x_new - xj) < self.eps)
                           for xj in self.x_vals]):
                    x_new[:] = self.lb[:] + (np.random.random(self.n)
                                             * (self.ub[:] - self.lb[:]))
        # Return the point to be sampled in a 2d array.
        return np.asarray([x_new])

    def save(self, filename):
        """ Save important data from this class so that it can be reloaded.

        Note: If this function is left unimplemented, ParMOO will reinitialize
        a fresh instance after a save/load. If this is the desired behavior,
        then this method and the load method need not be implemented.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data should be saved.

        """

        raise NotImplementedError("This class method has not been implemented")

    def load(self, filename):
        """ Reload important data into this class after a previous save.

        Note: If this function is left unimplemented, ParMOO will reinitialize
        a fresh instance after a save/load. If this is the desired behavior,
        then this method and the save method need not be implemented.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data has been saved.

        """

        raise NotImplementedError("This class method has not been implemented")


class SurrogateOptimizer(ABC):
    """ ABC describing surrogate optimization techniques.

    This class contains the following methods.
     * ``setObjective(obj_func)`` (default implementation provided)
     * ``setSimulation(sim_func, sd_func)`` (default implementation provided)
     * ``setConstraints(constraint_func)`` (default implementation provided)
     * ``setPenalty(penaltyFunc, gradFunc)`` (default implementation provided)
     * ``setTrFunc(trFunc)`` (default implementation provided)
     * ``addAcquisition(*args)`` (default implementation provided)
     * ``returnResults(x, fx, sx, sdx)``
     * ``solve(x)``
     * ``save(filename)``
     * ``load(filename)``

    """

    @abstractmethod
    def __init__(self, o, lb, ub, hyperparams):
        """ Constructor for the SurrogateOptimizer class.

        Args:
            o (int): The number of objectives.

            lb (ndarray): A 1D array of lower bounds for the design space.

            ub (ndarray): A 1D array of upper bounds for the design space.

            hyperparams (dict): A dictionary of hyperparameters for the
                optimization procedure.

        Returns:
            SurrogateOptimizer: A new SurrogateOptimizer object.

        """

    def setObjective(self, obj_func):
        """ Add a vector-valued objective function that will be solved.

        Args:
            obj_func (function): A vector-valued function that can be evaluated
                to solve the surrogate optimization problem.

        """

        # Check whether obj_func() has an appropriate signature
        if callable(obj_func):
            if len(inspect.signature(obj_func).parameters) != 2:
                raise ValueError("obj_func() must accept exactly two inputs")
            else:
                # Add obj_func to the problem
                self.objectives = obj_func
        else:
            raise TypeError("obj_func() must be callable")
        return

    def setSimulation(self, sim_func, sd_func):
        """ Add a vector-valued simulation function, used to calculate objs.

        Args:
            sim_func (function): A vector-valued function that can be evaluated
                to determine the surrogate-predicted simulation outputs.

            sd_func (function): A vector-valued function that can be evaluated
                to determine the standard deviations of the surrogate
                predictions.

        """

        # Check whether sim_func() has an appropriate signature
        if callable(sim_func):
            if len(inspect.signature(sim_func).parameters) != 1:
                raise ValueError("sim_func() must accept exactly one input")
            else:
                # Add sim_func to the problem
                self.simulations = sim_func
        else:
            raise TypeError("sim_func() must be callable")
        # Check whether sd_func() has an appropriate signature
        if callable(sd_func):
            if len(inspect.signature(sd_func).parameters) != 1:
                raise ValueError("sd_func() must accept exactly one input")
            else:
                self.sim_sd = sd_func
        else:
            raise TypeError("sd_func() must be callable")
        return

    def setPenalty(self, penalty_func):
        """ Add a matrix-valued gradient function for obj_func.

        Args:
            penalty_func (function): A vector-valued penalized objective
                that incorporates a penalty for violating constraints.

            grad_func (function): A matrix-valued function that can be
                evaluated to obtain the Jacobian matrix for obj_func.

        """

        # Check whether penalty_func() has an appropriate signature
        if callable(penalty_func):
            if len(inspect.signature(penalty_func).parameters) != 2:
                raise ValueError("penalty_func must accept exactly two inputs")
            else:
                # Add penalty to the problem
                self.penalty_func = penalty_func
        else:
            raise TypeError("penalty_func must be callable")
        return

    def setConstraints(self, constraint_func):
        """ Add a constraint function that will be satisfied.

        Args:
            constraint_func (function): A vector-valued function from the
                design space whose components correspond to constraint
                violations. If the problem has only bound constraints, this
                function returns zeros.

        """

        # Check whether constraint_func() has an appropriate signature
        if callable(constraint_func):
            if len(inspect.signature(constraint_func).parameters) != 2:
                raise ValueError("constraint_func() must accept exactly two"
                                 + " input")
            else:
                # Add constraint_func to the problem
                self.constraints = constraint_func
        else:
            raise TypeError("constraint_func() must be callable")
        return

    def setTrFunc(self, trFunc):
        """ Add a TR setter function for alerting surrogates.

        Args:
            trFunc (function): A function with 2 inputs, which the optimizer
                must call prior to solving each surrogate optimization problem
                in order to set the trust-region center and radius.

        """

        # Check whether trFunc() has an appropriate signature
        if callable(trFunc):
            if len(inspect.signature(trFunc).parameters) != 2:
                raise ValueError("trFunc() must accept exactly 2 inputs")
            else:
                # Add obj_func to the problem
                self.setTR = trFunc
        else:
            raise TypeError("trFunc() must be callable")
        return

    def returnResults(self, x, fx, sx, sdx):
        """ This is a callback function to collect evaluation results.

        Implement this function to receive the results of each
        true simulation evaluation from the MOOP class at runtime.

        Args:
            x (ndarray): A 1D array with the design point evaluated.

            fx (ndarray): A 1D array with the objective function values at x.

            sx (ndarray): The simulation function values at x.

            sdx (ndarray): The standard deviation in the simulation prediction.

        """

        return

    def addAcquisition(self, *args):
        """ Add an acquisition function for the surrogate optimizer.

        Args:
            *args (AcquisitionFunction): Acquisition functions that are used
                to scalarize the list of objectives in order to solve the
                surrogate optimization problem.

        """

        # Check for illegal inputs
        if not all([isinstance(arg, AcquisitionFunction) for arg in args]):
            raise TypeError("Args must be instances of AcquisitionFunction")
        # Append all arguments to the acquisitions list
        for arg in args:
            self.acquisitions.append(arg)
        return

    @abstractmethod
    def solve(self, x_k):
        """ Solve the surrogate problem.

        You may assume that the following internal attributes are defined
        and contain callable definitions of the objective, constraint,
        penalty, and simulation (surrogate) functions, respectively:
         * ``self.objectives``,
         * ``self.constraints``,
         * ``self.penalty_func``, and
         * ``self.simulations``.

        Additionally, you may assume that:
         * ``self.acquisitions`` contains a list of one or more
           ``AcquisitionFunction`` object instances, each of whose
           ``acq.scalarize(f_vals, x_vals, s_vals_mean, s_vals_sd)``
           is set and ready to call; and
         * ``self.setTR(x, r)`` can be called to set a trust-region
           centered at ``x`` with radius ``r`` (and re-fit the surrogates
           accordingly).

        Note: If implementing your own solver, try to jit (or re-jit) any of
        the objective, constraint, penalty, simulation surrogate, and/or
        acquisition functions after each call to ``self.setTR``.
        Additionally, if provided by the user,
        the objectives, constraints, penalty, and acq.scalarize,
        functions should all be differentiable by importing and
        calling ``jax.jacrev()``.

        Args:
            x_k (ndarray): A 2D array containing a list of current iterates.

        Returns:
            ndarray: A 2D array matching the shape of x_k specifying x_{k+1}.

        """

    def save(self, filename):
        """ Save important data from this class so that it can be reloaded.

        Note: If this function is left unimplemented, ParMOO will reinitialize
        a fresh instance after a save/load. If this is the desired behavior,
        then this method and the load method need not be implemented.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data should be saved.

        """

        raise NotImplementedError("This class method has not been implemented")

    def load(self, filename):
        """ Reload important data into this class after a previous save.

        Note: If this function is left unimplemented, ParMOO will reinitialize
        a fresh instance after a save/load. If this is the desired behavior,
        then this method and the save method need not be implemented.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data has been saved.

        """

        raise NotImplementedError("This class method has not been implemented")



""" This module contains tools to generate interactive Plotly-based graphs.

The functions are:

 Display data in interactive browser plot
  * ``generate_scatter(moop)`` -- Generate interactive scatterplot matrix
  * ``generate_parallel(moop)`` -- Generate interactive parallel plot
  * ``generate_radar(moop)`` -- Generate interactive radar plot

"""

import plotly.express as px
import plotly.graph_objects as go
import logging
from .utilities import (
    set_plot_name,
    set_database,
    set_hover_info,
)


def generate_scatter(
    moop,
    db,
    points,
):

    """ Generate a scatterplot or scatterplot matrix.

        Args:
            moop (MOOP): An object containing the data to be visualized.

            db (string): Filter traces by dataset.
             * 'pf' - Plot the Pareto Front.
             * 'obj' - Plot objective data.

            points (string): Filter traces by constraint score.
             * 'constraint_satisfying' - Show only points that
               satisfy every constraint.
             * 'constraint_violating' - Show only points that
               violate any constraint.
             * 'all' - Plot all points.
             * 'none' - Plot no points.

        Returns:
            plotly.graph_objects.Figure: A scatterplot or scatterplot matrix
            displaying traces that fit the filtering criteria.
    """

    # * intro log
    logging.info('generating scatterplot. this might take a while')

    # * get info
    objectives = moop.getObjectiveType().names

    # * choose database
    database = set_database(moop, db=db, points=points)
    plot_name = set_plot_name(db=db)

    # * create plot
    if (len(objectives) == 2):
        fig = px.scatter(
            database,
            x=objectives[0],
            y=objectives[1],
            title=plot_name,
            hover_data=database.columns,
            template='none',
        )
        # fig.update_xaxes(showticklabels=False)
        # fig.update_yaxes(showticklabels=False)
    else:
        fig = px.scatter_matrix(
            database,
            dimensions=objectives,
            title=plot_name,
            hover_data=database.columns,
            template='none',
        )
        fig.update_traces(diagonal_visible=False)

    # * logging outro
    logging.info('generated scatterplot')

    # * return figure
    return fig


def generate_parallel(
    moop,
    db,
    points,
):

    """ Generate a parallel coordinates plot.

        Args:
            moop (MOOP): An object containing the data to be visualized.

            db (string): Filter traces by dataset.
             * 'pf' - Plot the Pareto Front.
             * 'obj' - Plot objective data.

            points (string): Filter traces by constraint score.
             * 'constraint_satisfying' - Show only points that
               satisfy every constraint.
             * 'constraint_violating' - Show only points that
               violate any constraint.
             * 'all' - Plot all points.
             * 'none' - Plot no points.

        Returns:
            plotly.graph_objects.Figure: A parallel coordinates plot
            displaying traces that fit the filtering criteria.

    """

    # * intro log
    message = 'generating parallel coordinates plot. '
    message += 'this might take a while'
    logging.info(message)

    # * setup axes
    objectives = moop.getObjectiveType().names

    # * choose database
    database = set_database(moop, db=db, points=points)
    plot_name = set_plot_name(db=db)

    # * create plot
    fig = px.parallel_coordinates(
        database,
        dimensions=objectives,
        title=plot_name,
        template='none',
    )

    # * logging outro
    logging.info('generated parallel coordinates plot')

    # * return figure
    return fig


def generate_radar(
    moop,
    db,
    points,
):

    """ Generate a radar plot.

        Args:
            moop (MOOP): An object containing the data to be visualized.

            db (string): Filter traces by dataset.
             * 'pf' - Plot the Pareto Front.
             * 'obj' - Plot objective data.

            points (string): Filter traces by constraint score.
             * 'constraint_satisfying' - Show only points that
               satisfy every constraint.
             * 'constraint_violating' - Show only points that
               violate any constraint.
             * 'all' - Plot all points.
             * 'none' - Plot no points.

        Returns:
            plotly.graph_objects.Figure: A radar plot
            displaying traces that fit the filtering criteria.

    """

    # * intro log
    logging.info('generating radar plot. this might take awhile')

    # * setup axes
    objectives = moop.getObjectiveType().names
    wrap = objectives[0]
    temp_variable = list(objectives)
    temp_variable.append(wrap)
    axes = tuple(temp_variable)

    # * choose database
    database = set_database(moop, db=db, points=points)
    plot_name = set_plot_name(db=db)

    # * create scaled database
    j = database.copy(deep=True)
    for i in j.columns:
        j[i] = (j[i] - j[i].min()) / (j[i].max() - j[i].min())
    scaled_db = j

    # * create plot
    fig = go.Figure()
    for i in range(len(database)):
        trace = (i)
        hover_info = set_hover_info(database=database, i=i,)
        values = []
        for key in axes:
            values.append(scaled_db[key][i])
        fig.add_trace(go.Scatterpolar(
            r=values,
            theta=axes,
            name=trace,
            hovertext=hover_info,))

    # * improve aesthetics
    fig.update_layout(
        template='none',
    )
    fig.update_layout(
        polar=dict(
            radialaxis=dict(
                visible=True,
                showticklabels=False,)))
    fig.update_layout(
        title=dict(
            text=plot_name))
    fig.update_layout(
        autosize=True,)
    fig.update_layout(
        showlegend=False)

    # * logging outro
    logging.info('generated radar plot')

    # * return figure
    return fig




def test_UniformAugChebyshev():
    """ Test the UniformAugChebyshev class in acquisitions.py.

    Use the UniformAugChebyshev class to randomly sample 3 sets of convex
    weights from the unit simplex, and make sure that they are all convex.

    """

    from jax import jacrev
    import numpy as np
    from parmoo.acquisitions import UniformAugChebyshev
    import pytest

    # Initialize a good acquisition for future testing
    acqu = UniformAugChebyshev(3, np.zeros(4), np.ones(4), {})
    assert (np.all(acqu.lb[:] == 0.0) and np.all(acqu.ub[:] == 1.0))
    # Set some bad targets to check error handling
    with pytest.raises(TypeError):
        acqu.setTarget(5, lambda x: np.zeros(3))
    with pytest.raises(AttributeError):
        acqu.setTarget({'x_vals': []}, lambda x: np.zeros(3))
    with pytest.raises(ValueError):
        acqu.setTarget({'x_vals': np.ones(1), 'f_vals': np.ones(2)},
                       lambda x: np.zeros(3))
    with pytest.raises(ValueError):
        acqu.setTarget({'x_vals': np.ones((1, 1)), 'f_vals': np.ones((1, 3))},
                       lambda x: np.zeros(3))
    with pytest.raises(ValueError):
        acqu.setTarget({'x_vals': np.ones((1, 4)), 'f_vals': np.ones((1, 1))},
                       lambda x: np.zeros(3))
    with pytest.raises(TypeError):
        acqu.setTarget({}, 4)
    with pytest.raises(ValueError):
        acqu.setTarget({}, lambda x, y, z: np.zeros(3))
    # Set a good target for future usage
    x0 = acqu.setTarget({}, lambda x: np.zeros(3))
    assert (abs(sum(acqu.weights) - 1.0) < 1.0e-8)
    assert (np.all(x0[:] <= acqu.ub) and np.all(x0[:] >= acqu.lb))
    # Generate 3 random weight vector
    acqu1 = UniformAugChebyshev(3, np.zeros(4), np.ones(4), {})
    acqu1.setTarget({}, lambda x: np.zeros(3))
    acqu2 = UniformAugChebyshev(3, np.zeros(4), np.ones(4), {})
    acqu2.setTarget({}, lambda x: np.zeros(3))
    acqu3 = UniformAugChebyshev(3, np.zeros(4), np.ones(4), {})
    acqu3.setTarget({'x_vals': None, 'f_vals': None},
                    lambda x: np.zeros(3))
    # Check that the weights are all greater than 0 and sum to 1
    assert (all(acqu1.weights[:] >= 0.0))
    assert (all(acqu2.weights[:] >= 0.0))
    assert (all(acqu3.weights[:] >= 0.0))
    assert (abs(sum(acqu1.weights[:]) - 1.0) < 1.0e-8)
    assert (abs(sum(acqu2.weights[:]) - 1.0) < 1.0e-8)
    assert (abs(sum(acqu3.weights[:]) - 1.0) < 1.0e-8)
    # Check the scalarization and manifold checker appears to work correctly
    f_vals = np.random.sample(3)
    maxind = np.argmax(acqu1.weights * f_vals)
    assert (acqu1.getManifold(f_vals)[maxind] == 1)
    assert (abs(acqu1.scalarize(f_vals, np.ones(2), np.ones(2), np.ones(2)) -
                acqu1.weights[maxind] * f_vals[maxind]) < 3.0e-2)
    f_vals = np.random.sample(3)
    maxind = np.argmax(acqu2.weights * f_vals)
    assert (acqu2.getManifold(f_vals)[maxind] == 1)
    assert (abs(acqu2.scalarize(f_vals, np.ones(2), np.ones(2), np.ones(2)) -
                acqu2.weights[maxind] * f_vals[maxind]) < 3.0e-2)
    f_vals = np.random.sample(3)
    maxind = np.argmax(acqu3.weights * f_vals)
    assert (acqu3.getManifold(f_vals)[maxind] == 1)
    assert (abs(acqu3.scalarize(f_vals, np.ones(2), np.ones(2), np.ones(2)) -
                acqu3.weights[maxind] * f_vals[maxind]) < 3.0e-2)
    # Check the gradient scalarization appears to work correctly
    df1 = jacrev(acqu1.scalarize)(np.ones(3), np.ones(4),
                                  np.ones(1), np.ones(1))
    df2 = jacrev(acqu2.scalarize)(np.ones(3), np.ones(4),
                                  np.ones(1), np.ones(1))
    df3 = jacrev(acqu3.scalarize)(np.ones(3), np.ones(4),
                                  np.ones(1), np.ones(1))
    assert (np.abs(np.max(df1) - np.max(acqu1.weights) - 1.0e-2) < 1.0e-2)
    assert (np.abs(np.max(df2) - np.max(acqu2.weights) - 1.0e-2) < 1.0e-2)
    assert (np.abs(np.max(df3) - np.max(acqu3.weights) - 1.0e-2) < 1.0e-2)


def test_FixedAugChebyshev():
    """ Test the FixedAugChebyshev class in acquisitions.py.

    Use the FixedAugChebyshev class to try 2 sets of fixed convex weights
    from the unit simplex, and make sure that they are all convex.

    """

    from jax import jacrev
    import numpy as np
    from parmoo.acquisitions import FixedAugChebyshev
    import pytest

    # Try some bad initializations to test error handling
    with pytest.raises(TypeError):
        FixedAugChebyshev(3, np.zeros(4), np.ones(4), {'weights': 5.0})
    with pytest.raises(ValueError):
        FixedAugChebyshev(3, np.zeros(4), np.ones(4), {'weights': np.ones(2)})
    # Initialize a good acquisition for future testing
    acqu = FixedAugChebyshev(3, np.zeros(4), np.ones(4),
                             {'weights': np.ones((3))})
    assert (np.all(acqu.lb[:] == 0.0) and np.all(acqu.ub[:] == 1.0))
    acqu = FixedAugChebyshev(3, np.zeros(4), np.ones(4), {})
    assert (np.all(acqu.lb[:] == 0.0) and np.all(acqu.ub[:] == 1.0))
    assert (np.sum(acqu.weights[:]) - 1.0 < 1.0e-8)
    # Set some bad targets to check error handling
    with pytest.raises(TypeError):
        acqu.setTarget(5, lambda x: np.zeros(3))
    with pytest.raises(AttributeError):
        acqu.setTarget({'x_vals': []}, lambda x: np.zeros(3))
    with pytest.raises(ValueError):
        acqu.setTarget({'x_vals': np.ones(1), 'f_vals': np.ones(2)},
                       lambda x: np.zeros(3))
    with pytest.raises(ValueError):
        acqu.setTarget({'x_vals': np.ones((1, 1)), 'f_vals': np.ones((1, 3))},
                       lambda x: np.zeros(3))
    with pytest.raises(ValueError):
        acqu.setTarget({'x_vals': np.ones((1, 4)), 'f_vals': np.ones((1, 1))},
                       lambda x: np.zeros(3))
    with pytest.raises(TypeError):
        acqu.setTarget({}, 4)
    with pytest.raises(ValueError):
        acqu.setTarget({}, lambda x, y, z: np.zeros(3))
    # Set some good targets
    x0 = acqu.setTarget({'x_vals': None, 'f_vals': None},
                        lambda x: np.zeros(3))
    assert (np.all(x0[:] <= acqu.ub) and np.all(x0[:] >= acqu.lb))
    x0 = acqu.setTarget({'x_vals': np.zeros((1, 4)),
                         'f_vals': np.zeros((1, 3)),
                         'c_vals': np.zeros((1, 1))},
                        lambda x: np.zeros(3))
    assert (np.all(x0[:] <= acqu.ub) and np.all(x0[:] >= acqu.lb))
    x0 = acqu.setTarget({}, lambda x: np.zeros(3))
    assert (np.all(x0[:] <= acqu.ub) and np.all(x0[:] >= acqu.lb))
    # Use the scalarization function to check the weights sum to 1
    assert (np.abs(acqu.scalarize(np.eye(3)[0], np.ones(2),
                                  np.ones(2), np.ones(2))
                   + acqu.scalarize(np.eye(3)[1], np.ones(2),
                                    np.ones(2), np.ones(2))
                   + acqu.scalarize(np.eye(3)[2], np.ones(2),
                                    np.ones(2), np.ones(2)) - 1.0)
            < 1.1e-1)
    # Check the scalarization gradient
    da = jacrev(acqu.scalarize)(np.eye(3)[0], np.zeros(4),
                                np.zeros(1), np.zeros(1))
    assert (np.abs(np.sum(da) - acqu.weights[0]) < 1.1e-1)


if __name__ == "__main__":
    test_UniformAugChebyshev()
    test_FixedAugChebyshev()




""" Use ParMOO to solve the DTLZ2 problem, treating DTLZ2 as a simulation.

Uses unnamed variables, the dtlz2_sim simulation function, and the
single_sim_out objective functions to define the problem.

"""

import numpy as np
from parmoo import MOOP
from parmoo.optimizers import LocalSurrogate_BFGS
from parmoo.surrogates import GaussRBF
from parmoo.acquisitions import RandomConstraint
from parmoo.searches import LatinHypercube
from parmoo.simulations.dtlz import dtlz2_sim as sim_func
from parmoo.objectives import SingleSimObjective as obj_func
from parmoo.objectives import SingleSimGradient as obj_grad
from parmoo.constraints import SumOfSimSquaresBound as const_func
from parmoo.constraints import SumOfSimSquaresBoundGradient as const_grad

# Set the problem dimensions
NUM_DES = 3
NUM_OBJ = 3

# Create a MOOP
moop = MOOP(LocalSurrogate_BFGS)

# Add NUM_DES continuous design variables
for i in range(NUM_DES):
    moop.addDesign({'ub': 1.0, 'lb': 0.0,
                    'des_type': "continuous", 'des_tol': 1.0e-8})

# Add the simulation
moop.addSimulation({'name': "DTLZ2",
                    'm': NUM_OBJ,
                    'sim_func': sim_func(moop.getDesignType(),
                                         num_obj=NUM_OBJ,
                                         offset=0.6),
                    'search': LatinHypercube,
                    'surrogate': GaussRBF,
                    'hyperparams': {}})

# Add NUM_OBJ objective functions
for i in range(NUM_OBJ):
    moop.addObjective({'name': f"f{i+1}",
                       'obj_func': obj_func(moop.getDesignType(),
                                            moop.getSimulationType(),
                                            ("DTLZ2", i), goal="min"),
                       'obj_grad': obj_grad(moop.getDesignType(),
                                            moop.getSimulationType(),
                                            ("DTLZ2", i), goal="min")})

# Add a constraint
moop.addConstraint({'name': "SOS Sim Bounds",
                    'con_func': const_func(moop.getDesignType(),
                                           moop.getSimulationType(),
                                           sim_inds=[("DTLZ2", 0),
                                                     ("DTLZ2", 1),
                                                     ("DTLZ2", 2)],
                                           bound_type="upper",
                                           bound=4.0),
                    'con_grad': const_grad(moop.getDesignType(),
                                           moop.getSimulationType(),
                                           sim_inds=[("DTLZ2", 0),
                                                     ("DTLZ2", 1),
                                                     ("DTLZ2", 2)],
                                           bound_type="upper",
                                           bound=4.0)})

# Add 5 acquisition functions
for i in range(5):
    moop.addAcquisition({'acquisition': RandomConstraint, 'hyperparams': {}})

# Solve the problem with the equivalent of 5 iterations
moop.solve(sim_max=125)

# Check that 150 simulations were evaluated
assert(moop.getObjectiveData().shape[0] == 125)
assert(moop.getSimulationData()['DTLZ2'].shape[0] == 125)
assert(all([fi['f1']**2 + fi['f2']**2 + fi['f3']**2 <= 4
            for fi in moop.getPF()]))




""" A plotting library for interactive visualization of MOOP objects.

To generate a plot, generate create a MOOP object and use it to solve a
problem. Then call one of the following functions, passing the MOOP object
as the first argument:

 * ``viz.scatter(moop)``
 * ``viz.parallel_coordinates(moop)``
 * ``viz.radar(moop)``

Please take note of the following:

 * Via the interactivity features, all of the arguments except for the
   ``moop`` field can be adjusted interactively through the pop-up GUI.
   The plot type can also be adjusted from the dropdown menu while running
   the GUI.
 * The viz tool hosts a Dash app through the local port
   ``http://127.0.0.1:8050/``.
   This means that only one visualization can be hosted at a time.
   Also, killing the Dash app will end the interactivity in the browser,
   although a static plot may remain in your browser window.
 * Finally, note that there is a known issue when using the Chrome browser.
   The Firefox or Safari browsers are recommended.

To interact with the plots:

 1. To multi-select in scatterplots and radar plots, hold down ``SHIFT``
    while making additional selections.
 2. To remove all selections from a scatterplot and radar plot, select without
    holding down the ``SHIFT`` key or double click.
 3. To select in parallel coordinates plots, click along an axis and then
    drag the cursor elsewhere along the same axis.
    The highlighted part of the axis is the area that is selected.
    There is no need to hold down SHIFT when making multiple selections in
    parallel coordinates plots.
 4. To remove a parallel coordinates plot selection, click on the highlighted
    selection bar you wish to delete.
 5. To reset plot interactions made through the toolbar in the top right of a
    graph (not interactions made through buttons, dropdowns, toggles, or input
    boxes), double click on the plot. This will undo selections, zoom, pan,
    etc.

The three basic plot options are detailed below.

"""

import logging
from .dashboard import (
    Dash_App
)
from .utilities import (
    export_file,
    set_plot_name,
    check_inputs,
)
from .graph import (
    generate_scatter,
    generate_parallel,
    generate_radar,
)


def scatter(moop,
            db='pf',
            output='dash',
            points='constraint_satisfying',
            height='auto',
            width='auto',
            font='auto',
            fontsize='auto',
            background_color='auto',
            screenshot='svg',
            image_export_format='svg',
            data_export_format='csv',
            dev_mode=False,
            pop_up=True,
            port='http://127.0.0.1:8050/',
            ):
    """ Create a scatter plot matrix to visualize the results of a MOOP.

    Args:
        moop (MOOP): The MOOP results that you would like to visualize.

        db (str): Either 'pf' to plot just the Pareto front, or 'obj'
            to plot the complete objective database. Defaults to 'pf'.

        output (str): Either 'dash' to generate an interactive plot
            running in your browser using the dash app, or anything
            else to save a static plot to the desktop. Defaults to
            'dash'.

        points (str): Plot only
            constraint satisfying points ('constraint_satisfying'),
            constraint violating points ('constraint_violating'),
            all points ('all'),
            or no points ('none').

        height (str): The height in pixels of the resulting figure.
            Defaults to 'auto', which matches your screen size.

        width (str): The width in pixels of the resulting figure.
            Defaults to 'auto', which matches your screen size.

        font (str): The font that will be used for axis labels and legends.
            These values are automatically inferred from the name fields
            of your MOOP object. Any specified font must be available on your
            computer and available in the appropriate path. Defaults 'auto',
            which is times new roman on most machines.

        fontsize (str): The font size (in points). Defaults to 'auto', which
            infers the size based on the plot dimensions.

        background_color (str): Set the background color for this plot.
            Defaults to 'auto', which is white with grey axis lines on
            most systems.

        screenshot (str): Set the download mode when saving a screenshot
            using the "screenshot" button. Defaults to 'svg'.
            Other available options include:
            'html', 'webp', 'jpeg', 'png', 'svg', 'eps', and 'pdf'.
            Note that the 'eps' option requires the poppler library, which
            is not included in any of ParMOO's dependency lists.

        image_export_format (str): Set the export format when exporting
            a plot directly image file. Defaults to 'svg'.
            Other available options include:
            'html', 'webp', 'jpeg', 'png', 'svg', 'eps', and 'pdf'.
            Note that the 'eps' option requires the poppler library, which
            is not included in any of ParMOO's dependency lists.

        data_export_format (str): Set the format for exporting selected data
            to a file. Defaults to 'csv'. The other option is 'json'.

        dev_mode (bool): Run in developer mode, which allows changes to the
            code to automatically render in the browser. Activating this
            mode will interfere with some functionalities (such as
            checkpointing) since it results in multiple calls to the script.
            This value defaults to False, and should only be adjusted by
            developers.

        pop_up (bool): Automatically pop-up the dash app when called.
            Defaults to True. The only reason one might want to adjust, is
            if the environment prevents pop-ups or a non default browser
            is desired.

        port (str): The port through which the Dash app is hosted.
            Defaults to 'http://127.0.0.1:8050/'.

    """

    check_inputs(
        db=db,
        output=output,
        points=points,
        height=height,
        width=width,
        font=font,
        fontsize=fontsize,
        background_color=background_color,
        screenshot=screenshot,
        image_export_format=image_export_format,
        data_export_format=data_export_format,
        dev_mode=dev_mode,
        pop_up=pop_up,
        port=port,
    )

    logging.info('initialized scatter() wrapper')

    # * output
    if output == 'dash':
        Dash_App(
            plot_type='scatter',
            moop=moop,
            db=db,
            points=points,
            height=height,
            width=width,
            font=font,
            fontsize=fontsize,
            background_color=background_color,
            screenshot=screenshot,
            image_export_format=image_export_format,
            data_export_format=data_export_format,
            dev_mode=dev_mode,
            pop_up=pop_up,
            port=port,
        )
    else:
        plot_name = set_plot_name(db=db)
        export_file(
            fig=generate_scatter(
                moop,
                db=db,
                points=points,
            ),
            plot_name=plot_name,
            file_type=output,
        )


def parallel_coordinates(moop,
                         db='pf',
                         output='dash',
                         points='constraint_satisfying',
                         height='auto',
                         width='auto',
                         font='auto',
                         fontsize='auto',
                         background_color='auto',
                         screenshot='svg',
                         image_export_format='svg',
                         data_export_format='csv',
                         dev_mode=False,
                         pop_up=True,
                         port='http://127.0.0.1:8050/'
                         ):
    """ Create a parallel coordinates plot to visualize the results of a MOOP.

    Args:
        moop (MOOP): The MOOP results that you would like to visualize.

        db (str): Either 'pf' to plot just the Pareto front, or 'obj'
            to plot the complete objective database. Defaults to 'pf'.

        output (str): Either 'dash' to generate an interactive plot
            running in your browser using the dash app, or anything
            else to save a static plot to the desktop. Defaults to
            'dash'.

        points (str): Plot only
            constraint satisfying points ('constraint_satisfying'),
            constraint violating points ('constraint_violating'),
            all points ('all'),
            or no points ('none').

        height (str): The height in pixels of the resulting figure.
            Defaults to 'auto', which matches your screen size.

        width (str): The width in pixels of the resulting figure.
            Defaults to 'auto', which matches your screen size.

        font (str): The font that will be used for axis labels and legends.
            These values are automatically inferred from the name fields
            of your MOOP object. Any specified font must be available on your
            computer and available in the appropriate path. Defaults 'auto',
            which is times new roman on most machines.

        fontsize (str): The font size (in points). Defaults to 'auto', which
            infers the size based on the plot dimensions.

        background_color (str): Set the background color for this plot.
            Defaults to 'auto', which is white with grey axis lines on
            most systems.

        screenshot (str): Set the download mode when saving a screenshot
            using the "screenshot" button. Defaults to 'svg'.
            Other available options include:
            'html', 'webp', 'jpeg', 'png', 'svg', 'eps', and 'pdf'.
            Note that the 'eps' option requires the poppler library, which
            is not included in any of ParMOO's dependency lists.

        image_export_format (str): Set the export format when exporting
            a plot directly image file. Defaults to 'svg'.
            Other available options include:
            'html', 'webp', 'jpeg', 'png', 'svg', 'eps', and 'pdf'.
            Note that the 'eps' option requires the poppler library, which
            is not included in any of ParMOO's dependency lists.

        data_export_format (str): Set the format for exporting selected data
            to a file. Defaults to 'csv'. The other option is 'json'.

        dev_mode (bool): Run in developer mode, which allows changes to the
            code to automatically render in the browser. Activating this
            mode will interfere with some functionalities (such as
            checkpointing) since it results in multiple calls to the script.
            This value defaults to False, and should only be adjusted by
            developers.

        pop_up (bool): Automatically pop-up the dash app when called.
            Defaults to True. The only reason one might want to adjust, is
            if the environment prevents pop-ups or a non default browser
            is desired.

        port (str): The port through which the Dash app is hosted.
            Defaults to 'http://127.0.0.1:8050/'.

    """

    check_inputs(
        db=db,
        output=output,
        points=points,
        height=height,
        width=width,
        font=font,
        fontsize=fontsize,
        background_color=background_color,
        screenshot=screenshot,
        image_export_format=image_export_format,
        data_export_format=data_export_format,
        dev_mode=dev_mode,
        pop_up=pop_up,
        port=port,
    )

    logging.info('initialized parallel_coordinates() wrapper')

    if output == 'dash':
        Dash_App(
            plot_type='parallel',
            moop=moop,
            db=db,
            points=points,
            height=height,
            width=width,
            font=font,
            fontsize=fontsize,
            background_color=background_color,
            screenshot=screenshot,
            image_export_format=image_export_format,
            data_export_format=data_export_format,
            dev_mode=dev_mode,
            pop_up=pop_up,
            port=port,
        )
    else:
        plot_name = set_plot_name(db=db)
        export_file(
            fig=generate_parallel(
                moop=moop,
                db=db,
                points=points,
            ),
            plot_name=plot_name,
            file_type=output
        )


def radar(
          moop,
          db='pf',
          output='dash',
          points='constraint_satisfying',
          height='auto',
          width='auto',
          font='auto',
          fontsize='auto',
          background_color='auto',
          screenshot='svg',
          image_export_format='svg',
          data_export_format='csv',
          dev_mode=False,
          pop_up=True,
          port='http://127.0.0.1:8050/',
         ):
    """ Create a radar plot to visualize the results of a MOOP.

    Args:
        moop (MOOP): The MOOP results that you would like to visualize.

        db (str): Either 'pf' to plot just the Pareto front, or 'obj'
            to plot the complete objective database. Defaults to 'pf'.

        output (str): Either 'dash' to generate an interactive plot
            running in your browser using the dash app, or anything
            else to save a static plot to the desktop. Defaults to
            'dash'.

        points (str): Plot only
            constraint satisfying points ('constraint_satisfying'),
            constraint violating points ('constraint_violating'),
            all points ('all'),
            or no points ('none').

        height (str): The height in pixels of the resulting figure.
            Defaults to 'auto', which matches your screen size.

        width (str): The width in pixels of the resulting figure.
            Defaults to 'auto', which matches your screen size.

        font (str): The font that will be used for axis labels and legends.
            These values are automatically inferred from the name fields
            of your MOOP object. Any specified font must be available on your
            computer and available in the appropriate path. Defaults 'auto',
            which is times new roman on most machines.

        fontsize (str): The font size (in points). Defaults to 'auto', which
            infers the size based on the plot dimensions.

        background_color (str): Set the background color for this plot.
            Defaults to 'auto', which is white with grey axis lines on
            most systems.

        screenshot (str): Set the download mode when saving a screenshot
            using the "screenshot" button. Defaults to 'svg'.
            Other available options include:
            'html', 'webp', 'jpeg', 'png', 'svg', 'eps', and 'pdf'.
            Note that the 'eps' option requires the poppler library, which
            is not included in any of ParMOO's dependency lists.

        image_export_format (str): Set the export format when exporting
            a plot directly image file. Defaults to 'svg'.
            Other available options include:
            'html', 'webp', 'jpeg', 'png', 'svg', 'eps', and 'pdf'.
            Note that the 'eps' option requires the poppler library, which
            is not included in any of ParMOO's dependency lists.

        data_export_format (str): Set the format for exporting selected data
            to a file. Defaults to 'csv'. The other option is 'json'.

        dev_mode (bool): Run in developer mode, which allows changes to the
            code to automatically render in the browser. Activating this
            mode will interfere with some functionalities (such as
            checkpointing) since it results in multiple calls to the script.
            This value defaults to False, and should only be adjusted by
            developers.

        pop_up (bool): Automatically pop-up the dash app when called.
            Defaults to True. The only reason one might want to adjust, is
            if the environment prevents pop-ups or a non default browser
            is desired.

        port (str): The port through which the Dash app is hosted.
            Defaults to 'http://127.0.0.1:8050/'.

    """

    check_inputs(
        db=db,
        output=output,
        points=points,
        height=height,
        width=width,
        font=font,
        fontsize=fontsize,
        background_color=background_color,
        screenshot=screenshot,
        image_export_format=image_export_format,
        data_export_format=data_export_format,
        dev_mode=dev_mode,
        pop_up=pop_up,
        port=port,
    )
    logging.info('initialized radar() wrapper')
    if output == 'dash':
        Dash_App(
            plot_type='radar',
            moop=moop,
            db=db,
            points=points,
            height=height,
            width=width,
            font=font,
            fontsize=fontsize,
            background_color=background_color,
            screenshot=screenshot,
            image_export_format=image_export_format,
            data_export_format=data_export_format,
            dev_mode=dev_mode,
            pop_up=pop_up,
            port=port,
        )
    else:
        plot_name = set_plot_name(db=db)
        export_file(
            fig=generate_radar(
                moop,
                db=db,
                points=points,
            ),
            plot_name=plot_name,
            file_type=output
        )




""" Optimizers based on random search (RS).

This module contains implementations of the SurrogateOptimizer ABC, which
are based on randomized search strategies.

Note that these strategies are all gradient-free, and therefore does not
require objective, constraint, or surrogate gradient methods to be defined.

The classes include:
 * ``GlobalSurrogate_RS`` -- optimize surrogates globally via RS

"""

import jax
from jax import numpy as jnp
import numpy as np
from parmoo.structs import SurrogateOptimizer
from parmoo.util import xerror


class GlobalSurrogate_RS(SurrogateOptimizer):
    """ Use randomized search to identify potentially efficient designs.

    Randomly search the design space and use the surrogate models to predict
    whether each search point is potentially Pareto optimal.

    """

    # Slots for the GlobalSurrogate_RS class
    __slots__ = ['n', 'o', 'lb', 'ub', 'acquisitions', 'constraints',
                 'objectives', 'budget', 'simulations', 'setTR',
                 'penalty_func', 'sim_sd', 'np_rng']

    def __init__(self, o, lb, ub, hyperparams):
        """ Constructor for the GlobalSurrogate_RS class.

        Args:
            o (int): The number of objectives.

            lb (numpy.ndarray): A 1d array of lower bounds for the design
                region. The number of design variables is inferred from the
                dimension of lb.

            ub (numpy.ndarray): A 1d array of upper bounds for the design
                region. The dimension must match ub.

            hyperparams (dict): A dictionary of hyperparameters for the
                optimization procedure. It may contain the following:
                 * opt_budget (int): The sample size (default 10,000)

        Returns:
            SurrogateOptimizer: A new SurrogateOptimizer object.

        """

        # Check inputs
        xerror(o=o, lb=lb, ub=ub, hyperparams=hyperparams)
        self.o = o
        self.n = lb.size
        self.lb = lb
        self.ub = ub
        # Check that the contents of hyperparams is legal
        if 'opt_budget' in hyperparams:
            if isinstance(hyperparams['opt_budget'], int):
                if hyperparams['opt_budget'] < 1:
                    raise ValueError("hyperparams['opt_budget'] "
                                     "must be positive")
                else:
                    self.budget = hyperparams['opt_budget']
            else:
                raise TypeError("hyperparams['opt_budget'] "
                                "must be an integer")
        else:
            self.budget = 10000
        # Check the hyperparameter dictionary for random generator
        if 'np_random_gen' in hyperparams:
            if isinstance(hyperparams['np_random_gen'], np.random.Generator):
                self.np_rng = hyperparams['np_random_gen']
            else:
                raise TypeError("When present, hyperparams['np_random_gen'] "
                                "must be an instance of the class "
                                "numpy.random.Generator")
        else:
            self.np_rng = np.random.default_rng()
        # Initialize the list of acquisition functions
        self.acquisitions = []
        return

    def solve(self, x):
        """ Solve the surrogate problem using random search.

        Args:
            x (np.ndarray): A 2d array containing a list of feasible
                design points used to warm start the search.

        Returns:
            np.ndarray: A 2d numpy.ndarray containing a list of potentially
            efficient design points that were found by the random search.

        """

        from parmoo.util import updatePF

        # Check that x is legal
        if self.n != x.shape[1]:
            raise ValueError("The columns of x must match n")
        elif len(self.acquisitions) != x.shape[0]:
            raise ValueError("The rows of x must match the number " +
                             "of acquisition functions")
        # Initialize the surrogates with an infinite trust region
        rad = np.ones(self.n) * np.inf
        self.setTR(x[0, :], rad)
        # Compile the penalty function
        try:
            pen_func = jax.jit(self.penalty_func)
            x0 = x[0, :]
            sx0 = self.simulations(x0)
            f0 = pen_func(x0, sx0)
        except BaseException:
            pen_func = self.penalty_func
        # Set the batch size
        batch_size = 1000
        # Initialize the database
        data = {'x_vals': np.zeros((batch_size, self.n)),
                'f_vals': np.zeros((batch_size, self.o)),
                'c_vals': np.zeros((batch_size, 0))}
        # Loop over batch size until k == budget
        k = 0
        nondom = {}
        while (k < self.budget):
            # Check how many new points to generate
            k_new = min(self.budget, k + batch_size) - k
            if k_new < batch_size:
                data['x_vals'] = np.zeros((k_new, self.n))
                data['f_vals'] = np.zeros((k_new, self.o))
                data['c_vals'] = np.zeros((k_new, 0))
            # Randomly generate k_new new points
            for i in range(k_new):
                xi = (self.np_rng.random(self.n) *
                      (self.ub[:] - self.lb[:]) + self.lb[:])
                data['x_vals'][i, :] = xi[:]
                sxi = self.simulations(xi)
                data['f_vals'][i, :] = pen_func(xi, sxi)
            # Update the PF
            nondom = updatePF(data, nondom)
            k += k_new
        # Extract results for each scalarization via random search
        results = []
        for iq, acq in enumerate(self.acquisitions):
            # Compile the scalarization function
            if acq.useSD():

                def _sca_func(fi, xi):
                    return acq.scalarize(fi, xi, self.simulations(xi),
                                         self.sim_sd(xi))

            else:

                def _sca_func(fi, xi):
                    return acq.scalarize(fi, xi, self.simulations(xi),
                                         jnp.zeros(sxi.size))

            try:
                sca_func = jax.jit(_sca_func)
                _ = sca_func(f0, x0)
            except BaseException:
                sca_func = _sca_func
            # Use acquisition functions to extract array of results
            f_vals = [sca_func(fi, xi) for fi, xi in zip(nondom['f_vals'],
                                                         nondom['x_vals'])]
            imin = np.argmin(np.asarray([f_vals]))
            results.append(nondom['x_vals'][imin, :])
        self.objectives = None
        self.constraints = None
        self.penalty_func = None
        return np.asarray(results)



from .pattern_search import GlobalSurrogate_PS, LocalSurrogate_PS
from .random_search import GlobalSurrogate_RS
from .lbfgsb import GlobalSurrogate_BFGS, LocalSurrogate_BFGS

# LocalGPS = LocalSurrogate_PS
# GlobalGPS = GlobalSurrogate_PS
# RandomSearch = GlobalSurrogate_RS
# TR_LBFGSB = LocalSurrogate_BFGS
# LBFGSB = GlobalSurrogate_BFGS



""" This module contains several auxiliary functions used throughout ParMOO.

These functions may also be of external interest. They are:
 * xerror(o, lb, ub, hyperparams)
 * check_names(name, des_schema, sim_schema, obj_schema, con_schema)
 * check_sims(n, arg1, arg2, ...)
 * lex_leq(a, b)
 * updatePF(data, nondom)
 * to_array(x, dtype)
 * from_array(x, dtype)

"""

from jax import numpy as jnp
import numpy as np
from parmoo import structs
import inspect


def xerror(o=1, lb=None, ub=None, hyperparams=None):
    """ Typecheck the input arguments for a class interface.

    Args:
        o (int): The number of objectives should be an int greater than or
            equal to 1.

        lb (np.ndarray): The lower bounds should be a 1d array.

        ub (np.ndarray): The upper bounds should be a 1d array with the same
            length as lb, and must satisfy lb[:] < ub[:].

        hyperparams (dict): The hyperparameters must be supplied in a
            dictionary.

    """

    # Assign default values, if needed
    if lb is None:
        lb = np.zeros(1)
    if ub is None:
        ub = np.ones(1)
    if hyperparams is None:
        hyperparams = {}
    # Check the objective count
    if isinstance(o, int):
        if o < 1:
            raise ValueError("o must be positive")
    else:
        raise TypeError("o must be an integer")
    # Check that the bounds are legal
    if not isinstance(lb, np.ndarray):
        raise TypeError("lb must be a numpy array")
    if not isinstance(ub, np.ndarray):
        raise TypeError("ub must be a numpy array")
    if np.size(ub) != np.size(lb):
        raise ValueError("the dimensions of lb and ub must match")
    if np.any(lb >= ub):
        raise ValueError("ub must be strictly greater than lb")
    # Check the hyperparams dict
    if not isinstance(hyperparams, dict):
        raise TypeError("hyperparams must be a dictionary")
    return


def check_names(name, *args):
    """ Typecheck the input arguments for a new variable name.

    Args:
        name (str): A str that could be used as the variable name.

        *args (list of tuples): 1 or more lists of existing variable names to
            check against in order to guarantee that name is unique.

    """

    if not isinstance(name, str):
        raise TypeError("Every variable name must be a string type")
    for arg in args:
        if any([name == ni[0] for ni in arg]):
            raise ValueError(f"The variable name {name} is already in use")
    return


def check_sims(n, *args):
    """ Check simulation dictionaries for bad input.

    Args:
        n (int): The dimension of the design space. Used for confirming
            any simulation databases provided in args.

        *args (dict): An unpacked array of dictionaries, each specifying
            one of the simulations. The following keys are used:
             * name (String, optional): The name of this simulation
               (defaults to "sim" + str(i), where i = 1, 2, 3, ... for
               the first, second, third, ... simulation added to the
               MOOP).
             * m (int): The number of outputs for this simulation.
             * sim_func (function): An implementation of the simulation
               function, mapping from R^n -> R^m. The interface should
               match:
               `sim_out = sim_func(x, der=False)`,
               where `der` is an optional argument specifying whether
               to take the derivative of the simulation. Unless
               otherwise specified by your solver, `der` is always
               omitted by ParMOO's internal structures, and need not
               be implemented.
             * search (GlobalSearch): A GlobalSearch object for performing
               the initial search over this simulation's design space.
             * surrogate (SurrogateFunction): A SurrogateFunction object
               specifying how this simulation's outputs will be modeled.
             * des_tol (float): The tolerance for this simulation's
               design space; a new design point that is closer than
               des_tol to a point that is already in this simulation's
               database will not be reevaluated.
             * hyperparams (dict): A dictionary of hyperparameters, which
               will be passed to the surrogate and search routines.
               Most notably, search_budget (int) can be specified
               here.
             * sim_db (dict, optional): A dictionary of previous
               simulation evaluations. When present, contains:
                * x_vals (np.ndarray): A 2d array of pre-evaluated
                  design points.
                * s_vals (np.ndarray): A 2d array of corresponding
                  simulation outputs.
                * g_vals (np.ndarray): A 3d array of corresponding
                  Jacobian values. This value is only needed
                  if the provided SurrogateFunction uses gradients.

    """

    # Iterate through args to check each sim
    s = 0
    m = 0
    for arg in args:
        if isinstance(arg, dict):
            if 'name' in arg:
                if not isinstance(arg['name'], str):
                    raise TypeError("sims[" + str(s) + "]['name']"
                                    + " must be a string when present")
            # Check the number of sim outputs
            if 'm' in arg:
                if isinstance(arg['m'], int):
                    if arg['m'] > 0:
                        m = arg['m']
                    else:
                        raise ValueError("sims[" + str(s) + "]['m']"
                                         + " must be greater than zero")
                else:
                    raise TypeError("sims[" + str(s) + "] : 'm'"
                                    + " must be an int")
            else:
                raise AttributeError("sims[" + str(s)
                                     + "] is missing the key 'm'")
            # Get the hyperparameter dict
            if 'hyperparams' in arg:
                if not isinstance(arg['hyperparams'], dict):
                    raise TypeError("sims[" + str(s)
                                    + "]: 'hyperparams'"
                                    + " key must be a dict when present")
            # Check the search technique
            if 'search' in arg:
                try:
                    assert isinstance(arg['search'](m, np.zeros(n),
                                                    np.ones(n), {}),
                                      structs.GlobalSearch)
                except BaseException:
                    raise TypeError("sims[" + str(s) + "]['search']"
                                    + " must be a derivative of the"
                                    + " GlobalSearch abstract class")
            else:
                raise AttributeError("sims[" + str(s) + "] is missing"
                                     + " the key 'search'")
            # Check the des_tol, if present
            if 'des_tol' in arg:
                if isinstance(arg['des_tol'], float):
                    if arg['des_tol'] <= 0.0:
                        raise ValueError("sims[" + str(s)
                                         + "]['des_tol']"
                                         + " must be greater than 0")
                else:
                    raise TypeError("sims[" + str(s)
                                    + "]['des_tol'] must"
                                    + " be a float")
            # Get the surrogate function
            if 'surrogate' in arg:
                try:
                    if not isinstance(arg['surrogate'](m, np.zeros(n),
                                                       np.ones(n), {}),
                                      structs.SurrogateFunction):
                        raise TypeError("sims[" + str(s) + "] :"
                                        + " 'surrogate' must be a"
                                        + " derivative of the"
                                        + " SurrogateFunction abstract"
                                        + " class")
                except BaseException:
                    raise TypeError("sims[" + str(s)
                                    + "]['surrogate']"
                                    + " must be a derivative of the"
                                    + " SurrogateFunction abstract class")
            else:
                raise AttributeError("sims[" + str(s) + "] is missing"
                                     + " the key 'surrogate'")
            # Get the simulation function
            if 'sim_func' in arg:
                if callable(arg['sim_func']):
                    if len(inspect.signature(arg['sim_func']).parameters) \
                       != 1 and \
                       len(inspect.signature(arg['sim_func']).parameters) \
                       != 2:
                        raise ValueError("sims[" + str(s) + "]["
                                         + "'sim_func'] must accept"
                                         + " one or two inputs")
                else:
                    raise TypeError("sims[" + str(s)
                                    + "]['sim_func']"
                                    + " must be callable")
            else:
                raise AttributeError("sims[" + str(s) + "] is missing"
                                     + " the key 'sim_func'")
            # Get the starting database, if present
            if 'sim_db' in arg:
                if isinstance(arg['sim_db'], dict):
                    if 'x_vals' in arg['sim_db'] and \
                       's_vals' in arg['sim_db']:
                        try:
                            # Cast arg['sim_db'] contents to np.ndarrays
                            xvals = np.asarray(arg['sim_db']['x_vals'])
                            svals = np.asarray(arg['sim_db']['s_vals'],
                                               dtype=np.float64)
                        except BaseException:
                            raise TypeError("sims[" + str(s)
                                            + "]['sim_db']"
                                            + "['x_vals'] or sims["
                                            + str(s) + "]['sim_db']"
                                            + "['s_vals'] could not be"
                                            + " cast as a numpy array")
                        # Check the resulting dimensions
                        if xvals.size != 0 and svals.size != 0:
                            if xvals.ndim > 1:
                                if xvals.shape[1] != n:
                                    raise ValueError("sims[" + str(s)
                                                     + "]['sim_db']['x_vals']"
                                                     + " does not have"
                                                     + " n cols per row")
                                elif xvals.shape[0] != svals.shape[0]:
                                    raise ValueError("sims[" + str(s)
                                                     + "]['sim_db']['x_vals']"
                                                     + " does not have same"
                                                     + " number of rows as"
                                                     + " sims[" + str(s)
                                                     + "]['sim_db']['s_vals']")
                            if svals.shape[1] != m:
                                raise ValueError("sims[" + str(s)
                                                 + "]['sim_db']['s_vals']"
                                                 + " does not have"
                                                 + " sims[" + str(s)
                                                 + "]['m'] cols per row")
                        elif xvals.size != svals.size:
                            raise ValueError("sims[" + str(s)
                                             + "]['sim_db']['x_vals']"
                                             + " cannot be empty when"
                                             + " sims[" + str(s)
                                             + "]['sim_db']['s_vals']"
                                             + " is nonempty, and vice"
                                             + " versa")
                    elif 'x_vals' in arg['sim_db'] or \
                         's_vals' in arg['sim_db']:
                        raise AttributeError("sims[" + str(s) + "] cannot"
                                             + " contain a sim_db with"
                                             + " 'x_vals' but not 's_vals'"
                                             + " or vice versa")
                else:
                    raise TypeError("sims[" + str(s) + "]['sim_db']"
                                    + " must be a dict")
            s += 1
        else:
            raise TypeError("sims[" + str(s) + "] is not a dict")
        return


def lex_leq(a, b):
    """ Lexicographically compare two vectors from back to front.

    Check whether the vector a is lexicographically less than or equal to b,
    starting from the last element and working back to the first element.

    Args:
        a (numpy.ndarray): The first vector to compare.

        b (numpy.ndarray): The second vector to compare.

    Returns:
        Boolean: Whether a <= b in the lexicographical sense.

    """
    if a.size < 1 or b.size < 1:
        return True
    elif a[-1] < b[-1]:
        return True
    elif a[-1] > b[-1]:
        return False
    else:
        return lex_leq(a[:-1], b[:-1])


def updatePF(data, nondom):
    """ Update the Pareto front and efficient set by resorting.

    Returns:
        dict: A dictionary containing a discrete approximation of the
        Pareto front and efficient set.
         * f_vals (numpy.ndarray): A list of nondominated points
           discretely approximating the Pareto front.
         * x_vals (numpy.ndarray): A list of corresponding
           efficient design points.
         * c_vals (numpy.ndarray): A list of corresponding
           constraint satisfaction scores, all less than or equal to 0.

    """

    # Lexicographically sort all new points by 'f_vals'
    new_ind = np.lexsort(data['f_vals'][:, :].transpose())
    # Get problem size and allocate output array
    if 'f_vals' in nondom:
        nondom_len = nondom['f_vals'].shape[0]
        nondom_out = {'x_vals': np.zeros((nondom['x_vals'].shape[0]
                                          + data['x_vals'].shape[0],
                                          data['x_vals'].shape[1])),
                      'f_vals': np.zeros((nondom['f_vals'].shape[0]
                                          + data['f_vals'].shape[0],
                                          data['f_vals'].shape[1])),
                      'c_vals': np.zeros((nondom['c_vals'].shape[0]
                                          + data['c_vals'].shape[0],
                                          data['c_vals'].shape[1]))}
    else:
        nondom_len = 0
        nondom_out = {'x_vals': np.zeros((data['x_vals'].shape[0],
                                          data['x_vals'].shape[1])),
                      'f_vals': np.zeros((data['f_vals'].shape[0],
                                          data['f_vals'].shape[1])),
                      'c_vals': np.zeros((data['c_vals'].shape[0],
                                          data['c_vals'].shape[1]))}
    # Get total number of points to merge
    n_dat = nondom_len + data['f_vals'].shape[0]
    # Merge sorted lists
    j = 0
    k = 0
    for i in new_ind:
        # Add all points from nondom that are lexicographically first
        isNonDom = True
        while j < nondom_len and isNonDom:
            if lex_leq(nondom['f_vals'][j, :], data['f_vals'][i, :]):
                nondom_out['x_vals'][k, :] = nondom['x_vals'][j, :]
                nondom_out['f_vals'][k, :] = nondom['f_vals'][j, :]
                nondom_out['c_vals'][k, :] = nondom['c_vals'][j, :]
                j += 1
                k += 1
            else:
                isNonDom = False
        # Check for constraint violations
        if np.any(data['c_vals'][i, :] > 0.0):
            n_dat -= 1
        else:
            # If no constraints violated, append to the output
            nondom_out['x_vals'][k, :] = data['x_vals'][i, :]
            nondom_out['f_vals'][k, :] = data['f_vals'][i, :]
            nondom_out['c_vals'][k, :] = data['c_vals'][i, :]
            k += 1
    # Add remaining nondominated points from list
    while j < nondom_len:
        nondom_out['x_vals'][k, :] = nondom['x_vals'][j, :]
        nondom_out['f_vals'][k, :] = nondom['f_vals'][j, :]
        nondom_out['c_vals'][k, :] = nondom['c_vals'][j, :]
        j += 1
        k += 1
    # Loop over all points and look for nondominated points
    ndpts = 0  # counter for number of nondominated points
    for i in range(n_dat):
        # Check if data['f_vals'][i] is nondominated
        if np.all(np.any(nondom_out['f_vals'][i, :] <
                         nondom_out['f_vals'][:ndpts, :], axis=1)):
            # Swap entries at indices i and ndpts
            nondom_out['f_vals'][(i, ndpts), :] = \
                nondom_out['f_vals'][(ndpts, i), :]
            nondom_out['x_vals'][(i, ndpts), :] = \
                nondom_out['x_vals'][(ndpts, i), :]
            nondom_out['c_vals'][(i, ndpts), :] = \
                nondom_out['c_vals'][(ndpts, i), :]
            # Increment ndpts
            ndpts += 1
    # Return the solutions in a new dictionary
    return {'x_vals': nondom_out['x_vals'][:ndpts, :],
            'f_vals': nondom_out['f_vals'][:ndpts, :],
            'c_vals': nondom_out['c_vals'][:ndpts, :]}


def to_array(x, dtype):
    """ Unroll a ParMOO variable of the given dtype into a flat numpy array.

    Args:
        x (dict or numpy structured array): A ParMOO variable, which needs to
            be unrolled in a ndarray for convenience.

        dtype (numpy.dtype): The numpy.dtype of x.

    Returns:
        ndarray: a 1D array containing the values in x unrolled into an
        ndarray format, ordered by their order in the dtype.names.

    """

    xx = []
    for namei in dtype.names:
        xx.append(jnp.array([x[namei]]))
    return jnp.concatenate(xx, axis=None)


def from_array(x, dtype):
    """ Roll a flat ndarray back up into a ParMOO variable of the given dtype.

    Args:
        x (ndarray): A 1D array whose length matches the sum of the lengths of
            all fields for the dtype.

        dtype (numpy.dtype): The numpy.dtype of the needed output.

    Returns:
        dict: a ParMOO dictionary representation of x whose keys match the
        names in dtype and contains the values in x.

    """

    x1 = jnp.asarray(x).reshape((max(x.size, 1), ))
    xx = {}
    istart = 0
    for namei in dtype.names:
        if len(dtype[namei].shape) > 0:
            iend = istart + dtype[namei].shape[0]
        else:
            iend = istart + 1
        xx[namei] = x1[istart:iend].copy()
        istart = iend
    return xx



# Define a 5d problem with 3 objectives
n = 5
o = 3

def dtlz2_sim_unnamed(x):
    """ Evaluates the sim function for a collection of points given in
    ``H['x']``.

    """

    import math
    import numpy as np


    # Create output array for sim outs
    f = np.zeros(o)
    # Compute the kernel function g(x)
    gx = np.dot(x[o-1:n]-0.5, x[o-1:n]-0.5)
    # Compute the simulation outputs
    f[0] = (1.0 + gx)
    for y in x[:o-1]:
        f[0] *= math.cos(math.pi * y / 2.0)
    for i in range(1, o):
        f[i] = (1.0 + gx) * math.sin(math.pi * x[o-1-i] / 2.0)
        for y in x[:o-1-i]:
            f[i] *= math.cos(math.pi * y / 2.0)
    return f

# Define functions for named runs

def dtlz2_sim_named(x):
    """ Evaluates the sim function for a collection of points given in
    ``H['x']``.

    """

    import numpy as np

    # Unpack names into array
    xx = np.zeros(n)
    names = [f"x{i+1}" for i in range(n)]
    for i, name in enumerate(names):
        xx[i] = x[name]
    # Use dtlz2_sim to evaluate
    return dtlz2_sim_unnamed(xx)

def obj1_named(x, s): return s['DTLZ2'][0]
def obj2_named(x, s): return s['DTLZ2'][1]
def obj3_named(x, s): return s['DTLZ2'][2]




# @pytest.mark.extra
def test_libE_parmoo_persis_gen():
    """ Test the parmoo_persis_gen function in extras/libe.py.

    Generate several bad inputs to check for bad input.

    """

    import pytest

    try:
        from parmoo.extras.libe import parmoo_persis_gen
    except BaseException:
        pytest.skip("libEnsemble or its dependencies not importable. " +
                    "Skipping.")

    # Create libE dictionaries
    libE_info = {'comm': {}}
    persis_info = {'nworkers': 4}
    gen_specs = {'gen_f': parmoo_persis_gen,
                 'persis_in': ['x', 'sim_name', 'f'],
                 'out': [('x', float, 3),
                         ('sim_name', int)],
                 'user': {}}
    H = []

    # Try persis_gen with no persis_info['moop'] key
    with pytest.raises(KeyError):
        H, persis_info, exit_code = parmoo_persis_gen(H, persis_info,
                                                      gen_specs,
                                                      libE_info)
    # Try persis_gen with bad persis_info['moop'] key
    persis_info['moop'] = "hello world"
    with pytest.raises(TypeError):
        H, persis_info, exit_code = parmoo_persis_gen(H, persis_info,
                                                      gen_specs,
                                                      libE_info)


# @pytest.mark.extra
def test_libE_MOOP():
    """ Test the libE_MOOP class from extras/libe.py.

    Create a problem and check that the databases are empty.

    """

    import os
    import pytest
    from parmoo import MOOP
    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF
    from parmoo.acquisitions import RandomConstraint
    from parmoo.optimizers import LocalSurrogate_PS
    import numpy as np

    try:
        from parmoo.extras.libe import libE_MOOP
    except BaseException:
        pytest.skip("libEnsemble or its dependencies not importable. " +
                    "Skipping.")

    # Make all functions global for save/load
    global id_named, obj1, obj2, obj3, const1

    # Create a 5d problem with 3 objectives
    n = 5
    o = 3

    def id_named(x):
        """ Evaluates the sim function for a collection of points given in
        ``H['x']``.

        """

        # Unpack names into array
        xx = np.zeros(n)
        for i, name in enumerate(moop.moop.des_schema):
            xx[i] = x[name[0]]
        # Return ID
        return xx[:3]

    # Create a libE_MOOP with named variables
    moop = libE_MOOP(LocalSurrogate_PS)
    assert (isinstance(moop.moop, MOOP))
    moop = libE_MOOP(LocalSurrogate_PS, hyperparams={})
    assert (isinstance(moop.moop, MOOP))
    # Add n design vars
    for i in range(n):
        moop.addDesign({'name': "x" + str(i + 1), 'lb': 0.0, 'ub': 1.0})
    # Add simulation
    moop.addSimulation({'name': "Eye",
                        'm': o,
                        'sim_func': id_named,
                        'hyperparams': {'search_budget': 100},
                        'search': LatinHypercube,
                        'surrogate': GaussRBF,
                        'sim_db': {},
                        'des_tol': 0.00000001})
    # Add o objectives
    def obj1(x, s): return s['Eye'][0]
    def obj2(x, s): return s['Eye'][1]
    def obj3(x, s): return s['Eye'][2]
    moop.addObjective({'name': "obj1", 'obj_func': obj1})
    moop.addObjective({'name': "obj2", 'obj_func': obj2})
    moop.addObjective({'name': "obj3", 'obj_func': obj3})
    # Add 1 constraint
    def const1(x, s): return x["x5"] - 0.5
    moop.addConstraint({'name': "c1", 'constraint': const1})
    # Add 4 acquisition functions
    for i in range(4):
        moop.addAcquisition({'acquisition': RandomConstraint})
    moop.compile()
    assert (len(moop.getDesignType().names) == n)
    assert (len(moop.getSimulationType().names) == 1)
    assert (len(moop.getObjectiveType().names) == 3)
    assert (len(moop.getConstraintType().names) == 1)
    assert (len(moop.moop.acquisitions) == 4)
    # Perform 0 iteration manually
    batch = moop.iterate(0)
    for (xi, i) in batch:
        moop.evaluateSimulation(xi, i)
    moop.updateAll(0, batch)
    # Add a value in the simulation database
    x_val = {}
    for (name, _) in moop.moop.des_schema:
        x_val[name] = 0
    sx_val = {}
    for sname in moop.moop.sim_schema:
        if len(sname) > 2:
            sx_val[sname[0]] = np.zeros(sname[2])
        else:
            sx_val[sname[0]] = 0
    moop.updateSimDb(x_val, sx_val["Eye"], "Eye")
    assert (np.all(moop.checkSimDb(x_val, "Eye") == 0))
    moop.addObjData(x_val, sx_val)
    # Check Pareto front, objective data, sim data
    assert (moop.getPF()['x1'].shape[0] == 1)
    assert (moop.getObjectiveData()['x1'].shape[0] == 101)
    assert (moop.getSimulationData()['Eye']['x1'].shape[0] == 101)
    # Test checkpointing features
    moop.setCheckpoint(True)
    moop.save()
    moop.load()
    # Clean up test directory
    os.remove("parmoo.moop")
    os.remove("parmoo.surrogate.1")
    os.remove("parmoo.optimizer")


# @pytest.mark.extra
def test_libE_MOOP_bad_solve():
    """ Test the libE_MOOP.solve() method from extras/libe.py.

    Create a problem and check that the solve method handles bad input.

    """

    from parmoo.searches import LatinHypercube
    from parmoo.surrogates import GaussRBF
    from parmoo.acquisitions import RandomConstraint
    from parmoo.optimizers import LocalSurrogate_PS
    import pytest

    try:
        from parmoo.extras.libe import libE_MOOP
    except BaseException:
        pytest.skip("libEnsemble or its dependencies not importable. " +
                    "Skipping.")

    # Create a libE_MOOP
    moop = libE_MOOP(LocalSurrogate_PS)

    # Add 1 design var
    moop.addDesign({'lb': 0.0, 'ub': 1.0})

    # Add 1 simulation
    def id_named(x): return x[0]
    moop.addSimulation({'name': "Eye",
                        'm': 1,
                        'sim_func': id_named,
                        'hyperparams': {'search_budget': 100},
                        'search': LatinHypercube,
                        'surrogate': GaussRBF,
                        'sim_db': {},
                        'des_tol': 0.00000001})

    # Add 1 objective
    def obj1(x, s): return s[0]
    moop.addObjective({'obj_func': obj1})

    # Add 1 acquisition function
    moop.addAcquisition({'acquisition': RandomConstraint})

    # Hard code bad CL args
    import sys
    sys.argv.append("--comms")
    sys.argv.append("local")
    sys.argv.append("--nworkers")
    sys.argv.append("1")

    # Solve with unset budget vars
    with pytest.raises(ValueError):
        moop.solve()

    # Solve with bad type
    with pytest.raises(TypeError):
        moop.solve(iter_max="100")

    # Solve with bad CL args
    with pytest.raises(ValueError):
        moop.solve(iter_max=-1)

    # Solve with bad type
    with pytest.raises(TypeError):
        moop.solve(sim_max="100")

    # Solve with bad CL args
    with pytest.raises(ValueError):
        moop.solve(sim_max=-1)


if __name__ == "__main__":
    test_libE_parmoo_persis_gen()
    test_libE_MOOP()
    test_libE_MOOP_bad_solve()
